/**
 *	@file GLExtension.cpp
 *	@brief OpenGL extension parser implementation
 *	@author -tHE SWINe-
 *
 *	@date 2011-07-20
 *
 *	Added STATIC_AUTOCHECK_ALWAYS_IN_GLACHWRAP, EXPAND_GLAutoCheckFuncPtrName_MACRO
 *	and EXPAND_GLFuncPtrName_MACRO macros.
 *
 *	@date 2011-07-29
 *
 *	Added token name processing (so tokens of core extensions ending with _ARB are printed
 *	without it, and vice versa for non-core extensions).
 *
 *	Both tokens and functions are compared against and inserted to allocation tables
 *	after name processing (note this doesn't apply to function deprecation tables (todo?)).
 *
 *	Added code to generate documentation comments for OpenGL tokens and section \#ifdefs.
 */

#include "../UberLame_src/NewFix.h"
#include "../UberLame_src/CallStack.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <io.h>
#include <string>
#include <set>
#include <map>
#include <vector>
#include <stdexcept>
#include "../UberLame_src/Thread.h"
#include "../UberLame_src/MinMax.h"
#include "../UberLame_src/StlUtils.h"
#include "../UberLame_src/XML.h"
#include "../UberLame_src/Dir.h"
#include "../UberLame_src/UniConv.h"
#include "Download.h"
#include "GlExtension.h"

/**
 *	@def STATIC_AUTOCHECK_ALWAYS_IN_GLACHWRAP
 *	@brief if defined, GL_AUTOCHECK_BLOCK is always in *_glachwrap() function
 *		(old behavior) as opposed to having it in the do-while(0) block (the new
 *		behavior, enables the debugger display errors in user's source instead
 *		of in OpenGL40.h)
 *	@note Do not enable this macro, it generates worse headers with it.
 */
//#define STATIC_AUTOCHECK_ALWAYS_IN_GLACHWRAP

/**
 *	@def EXPAND_GLAutoCheckFuncPtrName_MACRO
 *	@brief if defined, the GLAutoCheckFuncPtrName() macro is expanded where possible;
 *		this doesn't incur any limitations and saves work to preprocessor when
 *		compiling OpenGL40.h, so it is generally a good thing to have it enabled
 */
#define EXPAND_GLAutoCheckFuncPtrName_MACRO

/**
 *	@def EXPAND_GLFuncPtrName_MACRO
 *	@brief if defined, the GLFuncPtrName() macro is expanded; this does have the
 *		limitation that FIX_GL_FUNCTION_LINKING_CONFLICTS must be enabled or disabled
 *		(based on the value of FIX_GL_FUNCTION_LINKING_CONFLICTS here), which
 *		may not be always the desired behavior (someone may want to export OpenGL
 *		functions and changing their names is bad then)
 */
//#define EXPAND_GLFuncPtrName_MACRO

/**
 *  @def FIX_GL_FUNCTION_LINKING_CONFLICTS
 *  @brief avoids function linking conflicts by renaming function pointers, and fixing
 *      names using #define
 *
 *  Some compilers may contain newer OpenGL libraries than this header requires
 *      (OpenGL 1.1, as supported by Microsoft Visual studio family), so names of function
 *      pointers, such as <tt>glBlendColor</tt> are present in static library, as well as in this
 *      file, causing link-time error. This is fixed by renaming the function, here
 *      <tt>glBlendColor_ptr</tt> would be used, and fixing it by macro, such as <tt>#define
 *      glBlendColor_ptr glBlendColor</tt>.
 */
#define FIX_GL_FUNCTION_LINKING_CONFLICTS

#ifdef FIX_GL_FUNCTION_LINKING_CONFLICTS
#define GLFuncPtrName_EXPANSION "_ptr"
#else // FIX_GL_FUNCTION_LINKING_CONFLICTS
#define GLFuncPtrName_EXPANSION ""
#endif // FIX_GL_FUNCTION_LINKING_CONFLICTS

#if !defined(__XML_LEXER_FORMAT_UTF_8) && !defined(__XML_LEXER_FORMAT_RAW_8)
#error("error: need 8-bit xml encoding")
#endif // !__XML_LEXER_FORMAT_UTF_8 && !__XML_LEXER_FORMAT_RAW_8
#if !defined(__XML_NODE_USE_DUMMY)
#error("error: need dummy xml node")
#endif // !__XML_NODE_USE_DUMMY
// check XML config

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif
// msvc 'for' scoping hack

#if defined(_WIN32) || defined(_WIN64)
#define strcasecmp(a,b) _stricmp((a),(b))
#endif // _WIN32 || _WIN64
// strcasecmp() for win32

#ifdef __XML_LEXER_FORMAT_UTF_8

static CUnicodeMapping some_8bit_mapping("./data/iso-8859-1.txt");

/**
 *	@brief converts string from utf-8 to iso-8859-1
 *	@param[in,out] r_s_string is the string to be converted
 */
static void DeUnicode(std::string &r_s_string)
{
	std::string s_temp;
	some_8bit_mapping.n_Set_SubsituteChar(' '); // !!
	CUniConv::n_Decode_UTF8(r_s_string.data(), r_s_string.length() * sizeof(char), s_temp, some_8bit_mapping, false);
	if(s_temp != r_s_string)
		s_temp.swap(r_s_string);
}

#else // __XML_LEXER_FORMAT_UTF_8

/**
 *	@brief converts string from utf-8 to iso-8859-1
 *	@param[in,out] r_s_string is the string to be converted
 */
static inline void DeUnicode(std::string &UNUSED(r_s_string))
{
	// nothing
}

#endif // __XML_LEXER_FORMAT_UTF_8

/**
 *	@brief removes leading and trailing whitespace
 *	@param[in,out] r_s_string is the string to have the whitespace removed
 */
void TrimSpace(std::string &r_s_string)
{
	int b = 0, e = r_s_string.length();
	while(e > 0 && (isspace((unsigned char)r_s_string[e - 1]) || (unsigned char)r_s_string[e - 1] == 0x00A0)) // 0x00A0 = no-break space
		-- e;
	while(b < e && (isspace((unsigned char)r_s_string[b]) || (unsigned char)r_s_string[e - 1] == 0x00A0))
		++ b;
	r_s_string.erase(e);
	r_s_string.erase(0, b);
}

/*
 *								=== CGLExtension ===
 */

CGLExtension::CGLExtension(const char *p_s_ext_name) // throws(std::bad_alloc)
	:m_s_ext_name(p_s_ext_name)
{
	if(strlen(p_s_ext_name) > 2 && strstr(p_s_ext_name, "GL_") == p_s_ext_name)
		p_s_ext_name += 3;
	// skip "GL_"

	const char *p_s_brand_begin = p_s_ext_name;
	while(*p_s_ext_name && *p_s_ext_name != '_')
		++ p_s_ext_name;
	const char *p_s_brand_end = p_s_ext_name;
	// get extension brand (such as ARB, EXT or NV)

	m_s_ext_brand.insert(m_s_ext_brand.begin(), p_s_brand_begin, p_s_brand_end);
	// makes extension brand

	Reset_FunctionParser();
}

bool CGLExtension::Parse_FromFile(const char *p_s_filename,
	const std::set<std::string> &r_deprecated_func_set, bool b_find_documentation,
	std::map<std::string, std::string> &r_fun_docu_url_map,
	const std::map<std::string, std::string> &r_fun_docu_alias_map)
{
	FILE *p_fr;
	if(!(p_fr = fopen(p_s_filename, "r"))) {
		fprintf(stderr, "error: failed to open \'%s\'\n", p_s_filename);
		return false;
	}
	// open file

	//printf("parsing \'%s\'\n", p_s_filename);
	// verbose

	bool b_had_procedures_section = false;
	bool b_had_tokens_section = false;
	bool b_proc_section_could_be_empty = false;
	bool b_proc_section_should_be_empty = false;
	bool b_tokens_section_could_be_empty = false;
	bool b_tokens_section_should_be_empty = false;

	int n_section = -1, n_line = 0;
	std::string s_line;
	while(!feof(p_fr)) {
		s_line.erase();
		for(int c = fgetc(p_fr); c != '\n' && c != EOF; c = fgetc(p_fr))
			s_line += c;
		if(!s_line.empty() && s_line[s_line.length() - 1] == '\r')
			s_line.erase(s_line.length() - 1); // don't want CR
		// read a line

		++ n_line;
		// maintain line counter

		if(!s_line.empty() && !isspace(s_line[0])) {
			if(!strcasecmp("New Procedures and Functions", s_line.c_str()) ||
			   s_line.find("New Procedures and Functions") == 0 ||
			   !strcasecmp("New Procedure and Functions", s_line.c_str()) || // ATI typo
			   s_line.find("New Procedure and Functions") == 0 ||
			   !strcasecmp("New Functions and Procedures", s_line.c_str()) || // ATI again
			   s_line.find("New Functions and Procedures") == 0) {
				//printf("new procedures and functions section begins on line %d\n", n_line); // verbose
				b_had_procedures_section = true;
				n_section = 1;
			} else if(!strcasecmp("New Tokens", s_line.c_str()) ||
			   s_line.find("New Tokens") == 0) {
				//printf("new tokens section begins on line %d\n", n_line); // verbose
				b_had_tokens_section = true;
				n_section = 2;
			} else {
				/*if(n_section != -1)
					printf("section ends on line %d\n", n_line);*/ // verbose
				n_section = -1;
				// @note GL_INTEL_texture_scissor has function headers starting at column 0, which are detected as a new section
				// @note similar thing happens in GL_SUN_vertex
				// @todo - parse and recognize all the section names to avoid this bug
			}
		}
		// find out what section we are in

		if(n_section == -1)
			continue;
		// skip code not in any important section

		if(s_line.find("None") != std::string::npos || s_line.find("none") != std::string::npos) {
			int b = 0, e = s_line.length();
			while(b < e && isspace(s_line[b]))
				++ b;
			_ASSERTE(b + 4 <= e);
			if((s_line[b + 0] == 'N' || s_line[b + 0] == 'n') && s_line[b + 1] == 'o' &&
			   s_line[b + 2] == 'n' && s_line[b + 3] == 'e') {
				if(n_section == 1)
					b_proc_section_could_be_empty = true;
				else
					b_tokens_section_could_be_empty = true;

				b += 4;
				while(b < e && (isspace(s_line[b]) || s_line[b] == '.' || s_line[b] == ','))
					++ b;
				if(b == e) {
					if(n_section == 1)
						b_proc_section_should_be_empty = true;
					else
						b_tokens_section_should_be_empty = true;
				}
			}
		}
		// match variations of "None" / "None." / "None ..." / "none" to indicate the section is empty

		if(n_section == 1) { // New Procedures and Functions
			bool b_has_proc_in_it = false;
			if(!b_FunctionParser_NeedsMore()) {
				for(size_t i = 0; i < sizeof(p_type_name_table) / sizeof(p_type_name_table[0]); ++ i) {
					size_t n_pos;
					if((n_pos = s_line.find(p_type_name_table[i].p_s_name)) != std::string::npos) {
						n_pos += strlen(p_type_name_table[i].p_s_name);
						// shift after the type (even though there might be longer types)

						size_t n_pos2;
						if((n_pos2 = s_line.find("(")) != std::string::npos && n_pos2 > n_pos) {
							b_has_proc_in_it = true;
							// maybe ...

							for(; n_pos < n_pos2; ++ n_pos) {
								if(s_line[n_pos] != '*' && s_line[n_pos] != '_' &&
								   s_line[n_pos] != '[' && s_line[n_pos] != '|' && s_line[n_pos] != ']' &&
								   s_line[n_pos] != '{' && s_line[n_pos] != '}' && !isspace(s_line[n_pos]) &&
								   !isdigit(s_line[n_pos]) && !isalpha(s_line[n_pos])) {
									b_has_proc_in_it = false;
									break;
								}
							}
							// check characters between data type, and 

							if(b_has_proc_in_it)
								break;
						}
					}
				}
				// try to find data type name, followed by a '('
				// @note GL_SGIX_list_priority has no return types (these are presumably void)
				// and the parser fails there, maybe there could be a way arround this
			}
			if(b_has_proc_in_it || b_FunctionParser_NeedsMore()) {
				if(!Parse_Procedure(s_line, p_s_filename, r_deprecated_func_set,
				   b_find_documentation, r_fun_docu_url_map, r_fun_docu_alias_map)) {
					Reset_FunctionParser();
					fprintf(stderr, "warning: Parse_Procedure() failed on line %d in \'%s\'\n", n_line, p_s_filename);
				}
			}
		} else if(n_section == 2) { // New Tokens
			TrimSpace(s_line); // !!
			if(s_line.find("0x") != std::string::npos ||
				(!s_line.empty() && isdigit(s_line[s_line.length() - 1]))) { // not-hexadecimal
				if(!Parse_Token(s_line, p_s_filename))
					fprintf(stderr, "warning: Parse_Token() failed on line %d in \'%s\'\n", n_line, p_s_filename);
			}
			// if there's 0x, there will probably be token definition
		}
	}

	if(b_FunctionParser_NeedsMore())
		fprintf(stderr, "error: function parser needs more data on line %d in \'%s\'\n", n_line, p_s_filename);
	// unexpected end of section? rubbish?

	if(!b_had_tokens_section)
		fprintf(stderr, "warning: no \'New Tokens\' section in \'%s\'\n", p_s_filename);
	else if(b_tokens_section_should_be_empty != b_NoNewTokens()) {
		if(b_tokens_section_should_be_empty) {
			fprintf(stderr, "warning: section \'New Tokens\' "
				"empty, but some tokens were parsed in \'%s\'\n", p_s_filename);
		} else {
			if(b_tokens_section_could_be_empty) {
				fprintf(stderr, "warning: section \'New Tokens\' "
					"is probably empty (strange wording), no tokens parsed in \'%s\'\n", p_s_filename);
			} else {
				fprintf(stderr, "warning: section \'New Tokens\' "
					"not empty, but no tokens parsed in \'%s\'\n", p_s_filename);
			}
		}
	}
	if(!b_had_procedures_section)
		fprintf(stderr, "warning: no \'New Procedures and Functions\' section in \'%s\'\n", p_s_filename);
	else if(b_proc_section_should_be_empty != b_NoNewProceduresOrFunctions()) {
		if(b_proc_section_should_be_empty) {
			fprintf(stderr, "warning: section \'New Procedures and Functions\' "
				"not empty, but some functions were parsed in \'%s\'\n", p_s_filename);
		} else {
			if(b_proc_section_could_be_empty) {
				fprintf(stderr, "warning: section \'New Procedures and Functions\' "
					"is probably empty (strange wording), no function parsed in \'%s\'\n", p_s_filename);
			} else {
				fprintf(stderr, "warning: section \'New Procedures and Functions\' "
					"not empty, but no function parsed in \'%s\'\n", p_s_filename);
			}
		}
	}
	// warn about missing sections, probably typo we didn't parse

	if(ferror(p_fr)) {
		fprintf(stderr, "error: i/o errors reading \'%s\'\n", p_s_filename);
		fclose(p_fr);
		return false;
	}
	fclose(p_fr);
	// close file

	return true;
}

bool CGLExtension::b_NoNewTokens() const
{
	return m_token_list.empty();
}

bool CGLExtension::b_NoNewProceduresOrFunctions() const
{
	return m_function_list.empty();
}

void CGLExtension::MarkAllocatedFunctions(std::set<std::string> &r_function_alloc,
	bool b_force_ext_brand_name, bool b_remove_ext_brand_name) const
{
	for(size_t i = 0, n = m_function_list.size(); i < n; ++ i) {
		const TFunction &f = m_function_list[i];
		if(r_function_alloc.find(p_s_Function_Name(f, b_force_ext_brand_name,
		   b_remove_ext_brand_name) + 2) == r_function_alloc.end()) {
			r_function_alloc.insert(p_s_Function_Name(f,
				b_force_ext_brand_name, b_remove_ext_brand_name) + 2);
		}
	}
}

void CGLExtension::MarkAllocatedTokens(std::set<std::string> &token_alloc,
	bool b_force_ext_brand_name, bool b_remove_ext_brand_name) const
{
	for(size_t i = 0, n = m_token_list.size(); i < n; ++ i) {
		const TToken &t = m_token_list[i];
		if(token_alloc.find(p_s_Token_Name(t, b_force_ext_brand_name,
		   b_remove_ext_brand_name) + 3) == token_alloc.end()) {
			token_alloc.insert(p_s_Token_Name(t, b_force_ext_brand_name,
				b_remove_ext_brand_name) + 3);
		}
	}
}

void CGLExtension::DumpTokens(FILE *p_fw, const std::set<std::string> &r_token_alloc,
	const char *p_s_section_comment, bool b_force_ext_brand_name, bool b_remove_ext_brand_name) const
{
	for(size_t i = 0, n = m_token_list.size(); i < n; ++ i) {
		const TToken &t = m_token_list[i];

		bool b_redefined = false;
		if(//r_token_alloc.find(t.s_name) != r_token_alloc.end() ||
		   r_token_alloc.find(p_s_Token_Name(t, b_force_ext_brand_name,
		   b_remove_ext_brand_name) + 3) != r_token_alloc.end()) {
			b_redefined = true;
			fprintf(p_fw, "//");
		}

		char p_s_fmt[64];
		if(t.n_value > INT_MAX) {
			if(t.n_value > UINT64_MAX)
				fprintf(stderr, "error: t.n_value > INT_MAX : unable to printf\n");
			if(t.b_decimal_value) {
				fprintf(p_fw, "#define GL_%-56s " PRIu64, p_s_Token_Name(t, b_force_ext_brand_name, b_remove_ext_brand_name) + 3, t.n_value);
				// note this can't keep n_value_digit_num
			} else {
				if(t.n_value_digit_num > 8) {
					sprintf(p_s_fmt, "#define GL_%%-56s 0x%%0%dx%%08xull", t.n_value_digit_num - 8);
					fprintf(p_fw, p_s_fmt, p_s_Token_Name(t, b_force_ext_brand_name, b_remove_ext_brand_name) + 3,
						uint32_t(t.n_value >> 32), uint32_t(t.n_value));
				} else {
					sprintf(p_s_fmt, "#define GL_%%-56s 0x%%0%dxu", t.n_value_digit_num);
					fprintf(p_fw, p_s_fmt, p_s_Token_Name(t, b_force_ext_brand_name, b_remove_ext_brand_name) + 3, uint32_t(t.n_value));
				}
				// tricky ;)
			}
		} else {
			sprintf(p_s_fmt, (t.b_decimal_value)? "#define GL_%%-56s %%0%dd" :
				"#define GL_%%-56s 0x%%0%dx", max(4, t.n_value_digit_num));
			fprintf(p_fw, p_s_fmt, p_s_Token_Name(t, b_force_ext_brand_name, b_remove_ext_brand_name) + 3, int(t.n_value));
		}

		if(b_redefined)
			fprintf(p_fw, " // (defined above)\n");
		else {
			if(t.s_comment.empty()) {
				if(p_s_section_comment && strlen(p_s_section_comment)) {
					fprintf(p_fw, " /**< @brief %s (%s) enum **/",
						m_s_ext_name.c_str(), p_s_section_comment);
				} else
					fprintf(p_fw, " /**< @brief %s enum **/", m_s_ext_name.c_str());
			} else {
				if(p_s_section_comment && strlen(p_s_section_comment)) {
					fprintf(p_fw, " /**< @brief %s (%s) enum (%s) **/",
						m_s_ext_name.c_str(), p_s_section_comment, t.s_comment.c_str());
				} else
					fprintf(p_fw, " /**< @brief %s enum (%s) **/", m_s_ext_name.c_str(), t.s_comment.c_str());
			}
			fprintf(p_fw, "\n");
		}
	}
	// dump tokens
}

void CGLExtension::Dump() const
{
	if(!m_token_list.empty()) {
		printf("New Tokens\n\n");

		DumpTokens(stdout, std::set<std::string>(), "", false, false);
	} else
		printf("No New Tokens\n");

	if(!m_function_list.empty()) {
		printf("\nNew Procedures And Functions\n\n");

		for(size_t i = 0, n = m_function_list.size(); i < n; ++ i) {
			const TFunction &f = m_function_list[i];
			printf("%s %s(", p_s_DataType_Name(f.t_return_type), f.s_name.c_str());
			for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
				const TParam &p = f.param_list[j];
				printf((p.s_name.empty())? "%s%s%s" : "%s%s %s", (j > 0)? ", " : "",
					p_s_DataType_Name(p.t_data_type), p.s_name.c_str());
			}
			printf(");\n");
		}
		// dump procedures
	} else
		printf("\nNo New Procedures Or Functions\n");
}

void CGLExtension::Print_Source(FILE *p_fw, const char *p_s_section_comment,
	bool b_force_ext_brand_name, bool b_remove_ext_brand_name,
	bool b_force_ext_brand_name_in_getprocaddress_string,
	bool b_remove_ext_brand_name_in_getprocaddress_string,
	const std::set<std::string> &r_function_alloc,
	const std::map<std::string, std::string> &r_ext_func_name_translation_table,
	const std::set<std::string> &r_deprecated_func_set,
	const std::map<std::string, std::string> &r_entry_point_name_translation_table) const
{
	const bool b_febn = b_force_ext_brand_name, b_rebn = b_remove_ext_brand_name;
	const bool b_febn2 = b_force_ext_brand_name_in_getprocaddress_string,
		b_rebn2 = b_remove_ext_brand_name_in_getprocaddress_string;
	// those are just very long and repeat all the time

	if(p_s_section_comment && strlen(p_s_section_comment)) {
		fprintf(p_fw, "\n/*\n"
						" *                              --- %s (%s) ---\n"
						" */\n\n", m_s_ext_name.c_str(), p_s_section_comment);
	} else {
		fprintf(p_fw, "\n/*\n"
						" *                              --- %s ---\n"
						" */\n\n", m_s_ext_name.c_str());
	}

	fprintf(p_fw, "bool GLEH_%s = false;\n\n", m_s_ext_name.c_str() + 3);

	if(!m_function_list.empty()) {
		fprintf(p_fw, "#ifdef __GENERATE_%s\n\n", m_s_ext_name.c_str());

		for(size_t i = 0, n = m_function_list.size(); i < n; ++ i) {
			const TFunction &f = m_function_list[i];

			bool b_deprecated = r_deprecated_func_set.find(f.s_name) != r_deprecated_func_set.end();
			bool b_redefinition = r_function_alloc.find(p_s_Function_Name(f, b_force_ext_brand_name,
				b_remove_ext_brand_name) + 2) != r_function_alloc.end();
			if(b_deprecated || b_redefinition)
				fprintf(p_fw, "//");

#ifndef EXPAND_GLFuncPtrName_MACRO
			fprintf(p_fw, "%s (GLApi *GLFuncPtrName(%s))(",
				p_s_DataType_Name(f.t_return_type), p_s_Function_Name(f, b_febn, b_rebn));
#else // EXPAND_GLFuncPtrName_MACRO
			fprintf(p_fw, "%s (GLApi *%s%s)(",
				p_s_DataType_Name(f.t_return_type), p_s_Function_Name(f, b_febn, b_rebn), GLFuncPtrName_EXPANSION);
#endif // EXPAND_GLFuncPtrName_MACRO

			for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
				const TParam &p = f.param_list[j];
				fprintf(p_fw, "%s%s", (j > 0)? ", " : "", p_s_DataType_Name(p.t_data_type));
			}
			fprintf(p_fw, ") INIT_FUNC_POINTERS;");

			if(b_deprecated)
				fprintf(p_fw, " // deprecated\n");
			else if(b_redefinition)
				fprintf(p_fw, " // defined above\n");
			else
				fprintf(p_fw, "\n");
		}
		// dump functions

		fprintf(p_fw, "\n#endif //__GENERATE_%s\n\n", m_s_ext_name.c_str());
	}

	std::string s_ext_name = m_s_ext_name;
	if(s_ext_name.length() >= 2 && s_ext_name[0] == 'G' && s_ext_name[1] == 'L')
		s_ext_name.erase(0, 2);
	for(size_t i = 0; i < s_ext_name.length(); ++ i) {
		if(i && s_ext_name[i - 1] == '_') {
			s_ext_name[i] = toupper(s_ext_name[i]);
			s_ext_name.erase(-- i, 1);
		}
	}
	// erase underspaces, make large letters

	s_ext_name.insert(0, "n_Get");
	s_ext_name += "FuncPointers";
	if(r_ext_func_name_translation_table.find(s_ext_name) !=
	   r_ext_func_name_translation_table.end())
		s_ext_name = (*r_ext_func_name_translation_table.find(s_ext_name)).second;
	// apply translation

	fprintf(p_fw, "int CGL4ExtensionHandler::%s()\n", s_ext_name.c_str());
	fprintf(p_fw, "{\n");
	fprintf(p_fw, "    int n_failed_functions = 0;\n\n");

	if(!m_function_list.empty()) {
		fprintf(p_fw, "#ifdef __GENERATE_%s\n\n", m_s_ext_name.c_str());

		for(size_t i = 0, n = m_function_list.size(); i < n; ++ i) {
			const TFunction &f = m_function_list[i];

			/*if(r_function_alloc.find(p_s_Function_Name(f, b_force_ext_brand_name,
			   b_remove_ext_brand_name) + 2) != r_function_alloc.end())
				fprintf(p_fw, "    //");
			else
				fprintf(p_fw, "    ");*/
			// do not comment these calls, the entry points may only be initialized
			// from this extension (not from the one that defined them) and these would then be 0

			bool b_deprecated = r_deprecated_func_set.find(f.s_name) != r_deprecated_func_set.end();
			if(b_deprecated)
				fprintf(p_fw, "    //");
			else
				fprintf(p_fw, "    ");
			// remove deprecated entry points

#ifndef EXPAND_GLFuncPtrName_MACRO
			fprintf(p_fw, "if(!(GLFuncPtrName(%s) = (%s(GLApi*)(",
				p_s_Function_Name(f, b_febn, b_rebn), p_s_DataType_Name(f.t_return_type));
#else // EXPAND_GLFuncPtrName_MACRO
			fprintf(p_fw, "if(!(%s%s = (%s(GLApi*)(",
				p_s_Function_Name(f, b_febn, b_rebn), GLFuncPtrName_EXPANSION, p_s_DataType_Name(f.t_return_type));
#endif // EXPAND_GLFuncPtrName_MACRO

			for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
				const TParam &p = f.param_list[j];
				fprintf(p_fw, "%s%s", (j > 0)? ", " : "", p_s_DataType_Name(p.t_data_type));
			}
			const char *p_s_entry_point_name = p_s_Function_Name(f, b_febn2, b_rebn2);
			if(r_entry_point_name_translation_table.find(p_s_entry_point_name) !=
			   r_entry_point_name_translation_table.end())
				p_s_entry_point_name = (*r_entry_point_name_translation_table.find(p_s_entry_point_name)).second.c_str();
			fprintf(p_fw, "))wglGetProcAddress(\"%s\"))) ++ n_failed_functions;\n", p_s_entry_point_name);
		}
		// dump functions

		fprintf(p_fw, "\n#endif //__GENERATE_%s\n\n", m_s_ext_name.c_str());
	} else
		fprintf(p_fw, "    // no new functions or procedures here\n\n");

	fprintf(p_fw, "    GLEH_%s = !n_failed_functions && b_SupportedExtension(\"%s\");\n\n",
		m_s_ext_name.c_str() + 3, m_s_ext_name.c_str());

	fprintf(p_fw, "    return n_failed_functions;\n");
	fprintf(p_fw, "}\n");

	fprintf(p_fw, "\n/*\n"
					" *                              --- ~%s ---\n"
					" */\n", m_s_ext_name.c_str());
}

void CGLExtension::Print_Header2(FILE *p_fw, const char *p_s_section_comment,
	std::map<std::string, std::string> &r_ext_func_name_translation_table) const
{
	std::string s_ext_name = m_s_ext_name;
	if(s_ext_name.length() >= 2 && s_ext_name[0] == 'G' && s_ext_name[1] == 'L')
		s_ext_name.erase(0, 2);
	for(size_t i = 0; i < s_ext_name.length(); ++ i) {
		if(i && s_ext_name[i - 1] == '_') {
			s_ext_name[i] = toupper(s_ext_name[i]);
			s_ext_name.erase(-- i, 1);
		}
	}
	// erase underspaces, make large letters

	s_ext_name.insert(0, "n_Get");
	s_ext_name += "FuncPointers";
	if(r_ext_func_name_translation_table.find(s_ext_name) !=
	   r_ext_func_name_translation_table.end())
		s_ext_name = (*r_ext_func_name_translation_table.find(s_ext_name)).second;
	// apply translation

	fprintf(p_fw, "\n    /**\n");
	if(!p_s_section_comment || !strlen(p_s_section_comment))
		fprintf(p_fw, "     *  @brief gets %s function pointers, and sets GLEH_%s\n",
			m_s_ext_name.c_str(), m_s_ext_name.c_str() + 3);
	else {
		fprintf(p_fw, "     *  @brief gets %s (%s) function pointers, and sets GLEH_%s\n",
			m_s_ext_name.c_str(), p_s_section_comment, m_s_ext_name.c_str() + 3);
	}
	fprintf(p_fw, "     *\n");
	fprintf(p_fw, "     *  @return Returns number of missing functions (zero means success).\n");
	if(m_function_list.empty()) {
		fprintf(p_fw, "     *\n");
		fprintf(p_fw, "     *  @note this extension has no functions, so this always returns zero.\n");
	}
	fprintf(p_fw, "     */\n");
	fprintf(p_fw, "    static int %s();\n", s_ext_name.c_str());
}

void CGLExtension::Print_Header(FILE *p_fw, bool b_want_autocheck, const char *p_s_section_comment,
	bool b_force_ext_brand_name, bool b_remove_ext_brand_name,
	const std::set<std::string> &r_token_alloc,
	const std::set<std::string> &r_function_alloc,
	const std::set<std::string> &r_deprecated_func_set) const
{
	const bool b_febn = b_force_ext_brand_name, b_rebn = b_remove_ext_brand_name;
	// those are just very long and repeat all the time

	if(p_s_section_comment && strlen(p_s_section_comment)) {
		fprintf(p_fw, "//                              --- %s (%s) ---\n\n", m_s_ext_name.c_str(), p_s_section_comment);
	} else {
		fprintf(p_fw, "//                              --- %s ---\n\n", m_s_ext_name.c_str());
	}

	fprintf(p_fw, "/**\n");
	fprintf(p_fw, " *  @def %s\n", m_s_ext_name.c_str());
	if(p_s_section_comment && strlen(p_s_section_comment)) {
		fprintf(p_fw, " *  @brief if defined, %s (%s) entry points are available\n",
			m_s_ext_name.c_str(), p_s_section_comment);
	} else
		fprintf(p_fw, " *  @brief if defined, %s entry points are available\n", m_s_ext_name.c_str());
	fprintf(p_fw, " */\n");

	fprintf(p_fw, "#ifndef %s\n", m_s_ext_name.c_str());
	fprintf(p_fw, "#define %s 1\n", m_s_ext_name.c_str());
	if(!m_function_list.empty()) {
		fprintf(p_fw, "\n/**\n");
		fprintf(p_fw, " *  @def __GENERATE_%s\n", m_s_ext_name.c_str());
		if(p_s_section_comment && strlen(p_s_section_comment)) {
			fprintf(p_fw, " *  @brief if defined, code to initialize %s (%s) entry points is compiled\n",
				m_s_ext_name.c_str(), p_s_section_comment);
		} else
			fprintf(p_fw, " *  @brief if defined, %s entry points are available\n", m_s_ext_name.c_str());
		fprintf(p_fw, " */\n");

		fprintf(p_fw, "#define __GENERATE_%s\n\n", m_s_ext_name.c_str());
	} else
		fprintf(p_fw, "\n");

	fprintf(p_fw, "extern bool GLEH_%s; /**< @brief set if %s is supported */\n\n",
		m_s_ext_name.c_str() + 3, m_s_ext_name.c_str());

	if(!m_token_list.empty()) {
		DumpTokens(p_fw, r_token_alloc, p_s_section_comment, b_force_ext_brand_name, b_remove_ext_brand_name);
		//fprintf(p_fw, "\n");
	} else
		fprintf(p_fw, "// no new tokens\n");
	// print tokens

	if(!m_function_list.empty()) {
		if(!b_Have_SomeDocumentation(m_function_list.front()))
			fprintf(p_fw, "\n");

		for(size_t i = 0, n = m_function_list.size(); i < n; ++ i) {
			const TFunction &f = m_function_list[i];

			bool b_deprecated = r_deprecated_func_set.find(f.s_name) != r_deprecated_func_set.end();
			bool b_redefinition = false;
			if(r_function_alloc.find(p_s_Function_Name(f, b_force_ext_brand_name,
			   b_remove_ext_brand_name) + 2) != r_function_alloc.end() || b_deprecated) {
				b_redefinition = true;
				fprintf(p_fw, "//");
			} else if(b_Have_SomeDocumentation(f))
				PrintFunctionDoc(f, p_fw);

#ifndef EXPAND_GLFuncPtrName_MACRO
			fprintf(p_fw, "extern %s (GLApi *GLFuncPtrName(%s))(",
				p_s_DataType_Name(f.t_return_type), p_s_Function_Name(f, b_febn, b_rebn));
#else // EXPAND_GLFuncPtrName_MACRO
			fprintf(p_fw, "extern %s (GLApi *%s%s)(",
				p_s_DataType_Name(f.t_return_type), p_s_Function_Name(f, b_febn, b_rebn), GLFuncPtrName_EXPANSION);
#endif // EXPAND_GLFuncPtrName_MACRO

			for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
				const TParam &p = f.param_list[j];
				_ASSERTE(!p.s_name.empty());
				fprintf(p_fw, "%s%s%s", (j > 0)? ", " : "",
					p_s_DataType_Name_Space(p.t_data_type), p.s_name.c_str());
			}
			fprintf(p_fw, ");");

			if(b_deprecated)
				fprintf(p_fw, " // deprecated\n");
			else if(b_redefinition)
				fprintf(p_fw, " // defined above\n");
			else
				fprintf(p_fw, "\n");
		}
		// dump functions

		bool b_newlined = false;
		for(size_t i = 0, n = m_function_list.size(); i < n; ++ i) {
			const TFunction &f = m_function_list[i];

			bool b_deprecated = r_deprecated_func_set.find(f.s_name) != r_deprecated_func_set.end();
			bool b_redefinition = r_function_alloc.find(p_s_Function_Name(f, b_force_ext_brand_name,
				b_remove_ext_brand_name) + 2) != r_function_alloc.end();

			if(b_want_autocheck) {
				if(f.t_return_type.n_type != type_Void || f.t_return_type.n_pointer || f.t_return_type.b_const) {
					if(!b_newlined) {
						b_newlined = true;
						fprintf(p_fw, "\n");
					}
					// newlines are hell

					if(b_deprecated || b_redefinition)
						fprintf(p_fw, "//");

					fprintf(p_fw, "inline %s%s_glachwrap(",
						p_s_DataType_Name_Space(f.t_return_type), p_s_Function_Name(f, b_febn, b_rebn));
					// this is the function, it always must be expanded

					for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
						const TParam &p = f.param_list[j];
						_ASSERTE(!p.s_name.empty());
						fprintf(p_fw, "%s%s%s", (j > 0)? ", " : "",
							p_s_DataType_Name_Space(p.t_data_type), p.s_name.c_str());
					}

#ifndef EXPAND_GLFuncPtrName_MACRO
					fprintf(p_fw, "%sconst char *p_s_filename, int n_line) { %sresult = GLFuncPtrName(%s)(",
						(f.param_list.empty())? "" : ", ", p_s_DataType_Name_Space(f.t_return_type),
						p_s_Function_Name(f, b_febn, b_rebn));
#else // EXPAND_GLFuncPtrName_MACRO
					fprintf(p_fw, "%sconst char *p_s_filename, int n_line) { %sresult = %s%s(",
						(f.param_list.empty())? "" : ", ", p_s_DataType_Name_Space(f.t_return_type),
						p_s_Function_Name(f, b_febn, b_rebn), GLFuncPtrName_EXPANSION);
#endif // EXPAND_GLFuncPtrName_MACRO

					for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
						const TParam &p = f.param_list[j];
						_ASSERTE(!p.s_name.empty());
						fprintf(p_fw, "%s%s", (j > 0)? "," : "", p.s_name.c_str());
					}
					fprintf(p_fw, "); GL_AUTOCHECK_BLOCK(\"%s\","
						" p_s_filename, n_line); return result; }\n", p_s_Function_Name(f, b_febn, b_rebn));
				}
				// non-void functions need inline *_glachwrap()
			}
		}
		// dump _glachwrap functions

		if(!b_Have_SomeDocumentation(m_function_list.front()))
			fprintf(p_fw, "\n");

		for(size_t i = 0, n = m_function_list.size(); i < n; ++ i) {
			const TFunction &f = m_function_list[i];

			bool b_deprecated = r_deprecated_func_set.find(f.s_name) != r_deprecated_func_set.end();
			bool b_redefinition = r_function_alloc.find(p_s_Function_Name(f, b_force_ext_brand_name,
				b_remove_ext_brand_name) + 2) != r_function_alloc.end();

			if(b_want_autocheck) {
				if(f.t_return_type.n_type != type_Void || f.t_return_type.n_pointer || f.t_return_type.b_const) {
					if(b_deprecated || b_redefinition)
						fprintf(p_fw, "//");
					else if(b_Have_SomeDocumentation(f))
						PrintFunctionDoc(f, p_fw, p_s_Function_Name(f, b_febn, b_rebn));
					fprintf(p_fw, "#define %s(", p_s_Function_Name(f, b_febn, b_rebn));
					for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
						const TParam &p = f.param_list[j];
						_ASSERTE(!p.s_name.empty());
						fprintf(p_fw, "%s%s", (j > 0)? "," : "", p.s_name.c_str());
					}

					fprintf(p_fw, ") GLAutoCheckFuncPtrName(%s)(", p_s_Function_Name(f, b_febn, b_rebn));
					// this cannot expand as it switches between the ordinary and _glachwrap function in debug / release

					for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
						const TParam &p = f.param_list[j];
						_ASSERTE(!p.s_name.empty());
						fprintf(p_fw, "%s%s", (j > 0)? ", " : "", p.s_name.c_str());
					}
					fprintf(p_fw, "%s)\n", (f.param_list.empty())? "GLAutoCheckExtraParamsSA" : " GLAutoCheckExtraParams");
					// #define to alias *_glachwrap() function
				} else {
					if(b_deprecated || b_redefinition)
						fprintf(p_fw, "//");
					else if(b_Have_SomeDocumentation(f))
						PrintFunctionDoc(f, p_fw, p_s_Function_Name(f, b_febn, b_rebn));
					fprintf(p_fw, "#define %s(", p_s_Function_Name(f, b_febn, b_rebn));
					for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
						const TParam &p = f.param_list[j];
						_ASSERTE(!p.s_name.empty());
						fprintf(p_fw, "%s%s", (j > 0)? "," : "", p.s_name.c_str());
					}

#ifndef EXPAND_GLFuncPtrName_MACRO
					fprintf(p_fw, ") do { GLFuncPtrName(%s)(", p_s_Function_Name(f, b_febn, b_rebn));
#else // EXPAND_GLFuncPtrName_MACRO
					fprintf(p_fw, ") do { %s%s(", p_s_Function_Name(f, b_febn, b_rebn), GLFuncPtrName_EXPANSION);
#endif // EXPAND_GLFuncPtrName_MACRO

					for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
						const TParam &p = f.param_list[j];
						_ASSERTE(!p.s_name.empty());
						fprintf(p_fw, "%s%s", (j > 0)? ", " : "", p.s_name.c_str());
					}
					fprintf(p_fw, "); GL_AUTOCHECK_BLOCK(\"%s\","
						" __FILE__, __LINE__); } while(0)\n", p_s_Function_Name(f, b_febn, b_rebn));
					// void functions can be replaced by an extremely simple macro
				}
				// write an autocheck function
			} else {
				if(b_deprecated || b_redefinition)
					fprintf(p_fw, "//");
				else if(b_Have_SomeDocumentation(f))
					PrintFunctionDoc(f, p_fw, p_s_Function_Name(f, b_febn, b_rebn));

				/*fprintf(p_fw, "#define %s GLFuncPtrName(%s)\n",
					p_s_Function_Name(f, b_febn, b_rebn), p_s_Function_Name(f, b_febn, b_rebn));*/
				// this is without args

				fprintf(p_fw, "#define %s(", p_s_Function_Name(f, b_febn, b_rebn));
				for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
					const TParam &p = f.param_list[j];
					_ASSERTE(!p.s_name.empty());
					fprintf(p_fw, "%s%s", (j > 0)? "," : "", p.s_name.c_str());
				}

#ifndef EXPAND_GLFuncPtrName_MACRO
				fprintf(p_fw, ") GLFuncPtrName(%s)(", p_s_Function_Name(f, b_febn, b_rebn));
#else // EXPAND_GLFuncPtrName_MACRO
				fprintf(p_fw, ") %s%s(", p_s_Function_Name(f, b_febn, b_rebn), GLFuncPtrName_EXPANSION);
#endif // EXPAND_GLFuncPtrName_MACRO

				for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
					const TParam &p = f.param_list[j];
					_ASSERTE(!p.s_name.empty());
					fprintf(p_fw, "%s%s", (j > 0)? ", " : "", p.s_name.c_str());
				}
				fprintf(p_fw, ")\n");
			}
		}
		// dump functions
	} else
		fprintf(p_fw, "\n// no new procedures or functions\n");
	// print functions

	fprintf(p_fw, "\n#endif // %s\n", m_s_ext_name.c_str());

	if(p_s_section_comment && strlen(p_s_section_comment)) {
		fprintf(p_fw, "\n//                              --- ~%s (%s) ---\n\n", m_s_ext_name.c_str(), p_s_section_comment);
	} else {
		fprintf(p_fw, "\n//                              --- ~%s ---\n\n", m_s_ext_name.c_str());
	}
}

void CGLExtension::Print_StaticHeader(FILE *p_fw, bool b_want_autocheck, const char *p_s_section_comment,
	bool b_force_ext_brand_name, bool b_remove_ext_brand_name,
	const std::set<std::string> &r_token_alloc,
	const std::set<std::string> &r_function_alloc,
	const std::set<std::string> &r_deprecated_func_set) const
{
	const bool b_febn = b_force_ext_brand_name, b_rebn = b_remove_ext_brand_name;
	// those are just very long and repeat all the time

	if(p_s_section_comment && strlen(p_s_section_comment)) {
		fprintf(p_fw, "//                              --- %s (%s) ---\n\n", m_s_ext_name.c_str(), p_s_section_comment);
	} else {
		fprintf(p_fw, "//                              --- %s ---\n\n", m_s_ext_name.c_str());
	}

	fprintf(p_fw, "/**\n");
	fprintf(p_fw, " *  @def %s\n", m_s_ext_name.c_str());
	if(p_s_section_comment && strlen(p_s_section_comment)) {
		fprintf(p_fw, " *  @brief if defined, %s (%s) entry points are available\n",
			m_s_ext_name.c_str(), p_s_section_comment);
	} else
		fprintf(p_fw, " *  @brief if defined, %s entry points are available\n", m_s_ext_name.c_str());
	fprintf(p_fw, " */\n");

	fprintf(p_fw, "#ifndef %s\n", m_s_ext_name.c_str());
	fprintf(p_fw, "#define %s 1\n", m_s_ext_name.c_str());
	if(!m_function_list.empty()) {
		fprintf(p_fw, "\n/**\n");
		fprintf(p_fw, " *  @def __GENERATE_%s\n", m_s_ext_name.c_str());
		if(p_s_section_comment && strlen(p_s_section_comment)) {
			fprintf(p_fw, " *  @brief if defined, code to initialize %s (%s) entry points is compiled\n",
				m_s_ext_name.c_str(), p_s_section_comment);
		} else
			fprintf(p_fw, " *  @brief if defined, %s entry points are available\n", m_s_ext_name.c_str());
		fprintf(p_fw, " */\n");

		fprintf(p_fw, "#define __GENERATE_%s\n\n", m_s_ext_name.c_str());
	} else
		fprintf(p_fw, "\n");

	fprintf(p_fw, "extern bool GLEH_%s; /**< @brief set if %s is supported */\n\n",
		m_s_ext_name.c_str() + 3, m_s_ext_name.c_str());

	if(!m_token_list.empty()) {
		DumpTokens(p_fw, r_token_alloc, p_s_section_comment, b_force_ext_brand_name, b_remove_ext_brand_name);
		fprintf(p_fw, "\n");
	} else
		fprintf(p_fw, "// no new tokens\n");
	// print tokens

	if(!m_function_list.empty()) {
		if(m_token_list.empty())
			fprintf(p_fw, "\n");

		fprintf(p_fw, "#ifdef __cplusplus\n");
		fprintf(p_fw, "extern \"C\" {\n");
		fprintf(p_fw, "#endif\n");

		for(size_t i = 0, n = m_function_list.size(); i < n; ++ i) {
			const TFunction &f = m_function_list[i];

			bool b_deprecated = r_deprecated_func_set.find(f.s_name) != r_deprecated_func_set.end();
			bool b_redefinition = r_function_alloc.find(p_s_Function_Name(f, b_force_ext_brand_name,
				b_remove_ext_brand_name) + 2) != r_function_alloc.end();

			if(b_deprecated || b_redefinition)
				fprintf(p_fw, "//");
			else if(b_Have_SomeDocumentation(f))
				PrintFunctionDoc(f, p_fw);
			// this is without args

			fprintf(p_fw, "GLStaticApi %sApiEntry %s(", p_s_DataType_Name_Space(f.t_return_type),
				p_s_Function_Name(f, b_febn, b_rebn));
			for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
				const TParam &p = f.param_list[j];
				_ASSERTE(!p.s_name.empty());
				fprintf(p_fw, "%s%s%s", (j > 0)? ", " : "",
					p_s_DataType_Name_Space(p.t_data_type), p.s_name.c_str());
			}
			fprintf(p_fw, ");");
			// print function header

			if(b_deprecated)
				fprintf(p_fw, " // deprecated\n");
			else if(b_redefinition)
				fprintf(p_fw, " // defined above\n");
			else
				fprintf(p_fw, "\n");
		}
		// dump functions

		fprintf(p_fw, "\n#ifdef __cplusplus\n");
		fprintf(p_fw, "}\n");
		fprintf(p_fw, "#endif\n");

		if(b_want_autocheck) {
			fprintf(p_fw, "\n#ifdef __HAVE_GL_AUTOCHECK\n\n");

#ifdef STATIC_AUTOCHECK_ALWAYS_IN_GLACHWRAP
			for(size_t i = 0, n = m_function_list.size(); i < n; ++ i) {
				const TFunction &f = m_function_list[i];

				bool b_deprecated = r_deprecated_func_set.find(f.s_name) != r_deprecated_func_set.end();
				bool b_redefinition = r_function_alloc.find(p_s_Function_Name(f, b_force_ext_brand_name,
					b_remove_ext_brand_name) + 2) != r_function_alloc.end();

				if(b_deprecated || b_redefinition)
					fprintf(p_fw, "//");

				fprintf(p_fw, "inline %s%s_glachwrap(",
					p_s_DataType_Name_Space(f.t_return_type), p_s_Function_Name(f, b_febn, b_rebn));
				// this is the function, it always must be expanded

				for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
					const TParam &p = f.param_list[j];
					_ASSERTE(!p.s_name.empty());
					fprintf(p_fw, "%s%s%s", (j > 0)? ", " : "",
						p_s_DataType_Name_Space(p.t_data_type), p.s_name.c_str());
				}
				if(f.t_return_type.n_type != type_Void || f.t_return_type.n_pointer || f.t_return_type.b_const) {
					fprintf(p_fw, "%sconst char *p_s_filename, int n_line) { %sresult = %s(",
						(f.param_list.empty())? "" : ", ", p_s_DataType_Name_Space(f.t_return_type),
						p_s_Function_Name(f, b_febn, b_rebn));
				} else {
					fprintf(p_fw, "%sconst char *p_s_filename, int n_line) { %s(",
						(f.param_list.empty())? "" : ", ", p_s_Function_Name(f, b_febn, b_rebn));
				}
				for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
					const TParam &p = f.param_list[j];
					_ASSERTE(!p.s_name.empty());
					fprintf(p_fw, "%s%s", (j > 0)? "," : "", p.s_name.c_str());
				}
				if(f.t_return_type.n_type != type_Void || f.t_return_type.n_pointer || f.t_return_type.b_const) {
					fprintf(p_fw, "); GL_AUTOCHECK_BLOCK(\"%s\","
						" p_s_filename, n_line); return result; }\n", p_s_Function_Name(f, b_febn, b_rebn));
				} else {
					fprintf(p_fw, "); GL_AUTOCHECK_BLOCK(\"%s\","
						" p_s_filename, n_line); }\n", p_s_Function_Name(f, b_febn, b_rebn));
				}
				// static functions need inline *_glachwrap()
			}
			
			for(size_t i = 0, n = m_function_list.size(); i < n; ++ i) {
				const TFunction &f = m_function_list[i];

				bool b_deprecated = r_deprecated_func_set.find(f.s_name) != r_deprecated_func_set.end();
				bool b_redefinition = r_function_alloc.find(p_s_Function_Name(f, b_force_ext_brand_name,
					b_remove_ext_brand_name) + 2) != r_function_alloc.end();

				if(b_deprecated || b_redefinition)
					fprintf(p_fw, "//");
				else if(b_Have_SomeDocumentation(f))
					PrintFunctionDoc(f, p_fw, p_s_Function_Name(f, b_febn, b_rebn));
				fprintf(p_fw, "#define %s(", p_s_Function_Name(f, b_febn, b_rebn));
				for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
					const TParam &p = f.param_list[j];
					_ASSERTE(!p.s_name.empty());
					fprintf(p_fw, "%s%s", (j > 0)? "," : "", p.s_name.c_str());
				}
#ifndef EXPAND_GLAutoCheckFuncPtrName_MACRO
				fprintf(p_fw, ") GLAutoCheckFuncPtrName(%s)(", p_s_Function_Name(f, b_febn, b_rebn));
#else // EXPAND_GLAutoCheckFuncPtrName_MACRO
				fprintf(p_fw, ") %s_glachwrap(", p_s_Function_Name(f, b_febn, b_rebn));
				// this can expand as it is in __HAVE_GL_AUTOCHECK conditional block
#endif // EXPAND_GLAutoCheckFuncPtrName_MACRO
				for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
					const TParam &p = f.param_list[j];
					_ASSERTE(!p.s_name.empty());
					fprintf(p_fw, "%s%s", (j > 0)? ", " : "", p.s_name.c_str());
				}
				fprintf(p_fw, "%s__FILE__, __LINE__)\n", (f.param_list.empty())? "" : ", ");
				// and one more #define to alias *_glachwrap() function
			}
#else // STATIC_AUTOCHECK_ALWAYS_IN_GLACHWRAP
			for(size_t i = 0, n = m_function_list.size(); i < n; ++ i) {
				const TFunction &f = m_function_list[i];

				bool b_deprecated = r_deprecated_func_set.find(f.s_name) != r_deprecated_func_set.end();
				bool b_redefinition = r_function_alloc.find(p_s_Function_Name(f, b_force_ext_brand_name,
					b_remove_ext_brand_name) + 2) != r_function_alloc.end();

				if(b_deprecated || b_redefinition)
					fprintf(p_fw, "//");

				fprintf(p_fw, "inline %s%s_glachwrap(",
					p_s_DataType_Name_Space(f.t_return_type), p_s_Function_Name(f, b_febn, b_rebn));
				// this is the function, it always must be expanded

				for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
					const TParam &p = f.param_list[j];
					_ASSERTE(!p.s_name.empty());
					fprintf(p_fw, "%s%s%s", (j > 0)? ", " : "",
						p_s_DataType_Name_Space(p.t_data_type), p.s_name.c_str());
				}
				if(f.t_return_type.n_type != type_Void || f.t_return_type.n_pointer || f.t_return_type.b_const) {
					fprintf(p_fw, "%sconst char *p_s_filename, int n_line) { %sresult = %s(",
						(f.param_list.empty())? "" : ", ", p_s_DataType_Name_Space(f.t_return_type),
						p_s_Function_Name(f, b_febn, b_rebn));
				} else {
					fprintf(p_fw, ") { %s(", p_s_Function_Name(f, b_febn, b_rebn));
					// no autocheck parameters in case it's not a void function
				}
				for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
					const TParam &p = f.param_list[j];
					_ASSERTE(!p.s_name.empty());
					fprintf(p_fw, "%s%s", (j > 0)? "," : "", p.s_name.c_str());
				}
				if(f.t_return_type.n_type != type_Void || f.t_return_type.n_pointer || f.t_return_type.b_const) {
					fprintf(p_fw, "); GL_AUTOCHECK_BLOCK(\"%s\","
						" p_s_filename, n_line); return result; }\n", p_s_Function_Name(f, b_febn, b_rebn));
				} else {
					fprintf(p_fw, "); }\n");
					// no autocheck block in case it's not a void function
				}
				// static functions need inline *_glachwrap()
			}
			
			for(size_t i = 0, n = m_function_list.size(); i < n; ++ i) {
				const TFunction &f = m_function_list[i];

				bool b_deprecated = r_deprecated_func_set.find(f.s_name) != r_deprecated_func_set.end();
				bool b_redefinition = r_function_alloc.find(p_s_Function_Name(f, b_force_ext_brand_name,
					b_remove_ext_brand_name) + 2) != r_function_alloc.end();

				if(b_deprecated || b_redefinition)
					fprintf(p_fw, "//");
				else if(b_Have_SomeDocumentation(f))
					PrintFunctionDoc(f, p_fw, p_s_Function_Name(f, b_febn, b_rebn));
				fprintf(p_fw, "#define %s(", p_s_Function_Name(f, b_febn, b_rebn));
				for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
					const TParam &p = f.param_list[j];
					_ASSERTE(!p.s_name.empty());
					fprintf(p_fw, "%s%s", (j > 0)? "," : "", p.s_name.c_str());
				}
#ifndef EXPAND_GLAutoCheckFuncPtrName_MACRO
				if(f.t_return_type.n_type != type_Void || f.t_return_type.n_pointer || f.t_return_type.b_const) {
					fprintf(p_fw, ") GLAutoCheckFuncPtrName(%s)(", p_s_Function_Name(f, b_febn, b_rebn));
					// just call autocheck function
				} else {
					fprintf(p_fw, ") do { GLAutoCheckFuncPtrName(%s)(", p_s_Function_Name(f, b_febn, b_rebn));
					// begin do-while(0) block and autocheck function
				}
#else // EXPAND_GLAutoCheckFuncPtrName_MACRO
				if(f.t_return_type.n_type != type_Void || f.t_return_type.n_pointer || f.t_return_type.b_const) {
					fprintf(p_fw, ") %s_glachwrap(", p_s_Function_Name(f, b_febn, b_rebn));
					// just call autocheck function
				} else {
					fprintf(p_fw, ") do { %s_glachwrap(", p_s_Function_Name(f, b_febn, b_rebn));
					// begin do-while(0) block and autocheck function
				}
				// this can expand as it is in __HAVE_GL_AUTOCHECK conditional block
#endif // EXPAND_GLAutoCheckFuncPtrName_MACRO
				for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
					const TParam &p = f.param_list[j];
					_ASSERTE(!p.s_name.empty());
					fprintf(p_fw, "%s%s", (j > 0)? ", " : "", p.s_name.c_str());
				}
				if(f.t_return_type.n_type != type_Void || f.t_return_type.n_pointer || f.t_return_type.b_const) {
					fprintf(p_fw, "%s__FILE__, __LINE__)\n", (f.param_list.empty())? "" : ", ");
					// append autocheck parameters
				} else {
					fprintf(p_fw, "); GL_AUTOCHECK_BLOCK(\"%s\","
						" __FILE__, __LINE__); } while(0)\n", p_s_Function_Name(f, b_febn, b_rebn));
					// inplace autocheck block
				}
				// and one more #define to alias *_glachwrap() function
			}
#endif // STATIC_AUTOCHECK_ALWAYS_IN_GLACHWRAP

			fprintf(p_fw, "\n#endif // __HAVE_GL_AUTOCHECK\n");
		} else
			fprintf(p_fw, "\n");
	} else
		fprintf(p_fw, "// no new procedures or functions\n");
	// print functions

	fprintf(p_fw, "\n#endif // %s\n", m_s_ext_name.c_str());

	if(p_s_section_comment && strlen(p_s_section_comment)) {
		fprintf(p_fw, "\n//                              --- ~%s (%s) ---\n\n", m_s_ext_name.c_str(), p_s_section_comment);
	} else {
		fprintf(p_fw, "\n//                              --- ~%s ---\n\n", m_s_ext_name.c_str());
	}
}

bool CGLExtension::Parse_Token(const std::string &r_s_token, const char *p_s_filename) // throws(std::bad_alloc)
{
	int b = 0, e = r_s_token.length();

	while(b < e && isspace(r_s_token[b]))
		++ b;
	// skip space

	bool b_have_to_generate_multiple_tokens = false;
	bool b_add_number_tag_to_value = true;
	if(r_s_token.find("<n>") != std::string::npos)
		b_have_to_generate_multiple_tokens = true;
	if(b_have_to_generate_multiple_tokens)
		fprintf(stderr, "warning: number tag present in token in \'%s\'\n", p_s_filename);
	// is the number tag present?

	std::vector<int> number_tag_position_list;
	int n_name_begin = b;
	for(;;) {
		while(b < e && b_IsTokenChar(r_s_token[b], false/*b == n_name_begin*/)) // token names may start with a number (eg. (GL_)1PASS_SGIS from GL_SGIS_multisample)
			++ b;
		if(!b_have_to_generate_multiple_tokens)
			break;
		if(b + 3 <= e && r_s_token[b] == '<' && tolower(r_s_token[b + 1]) == 'n' && r_s_token[b + 2] == '>') {
			number_tag_position_list.push_back(b);
			b += 3;
		} else
			break;
		// parse the number tag and continue
	}
	int n_name_end = b;
	// get token name

	if(b_have_to_generate_multiple_tokens && number_tag_position_list.empty())
		fprintf(stderr, "warning: number tag not present in token name in \'%s\'\n", p_s_filename);
	else if(b_have_to_generate_multiple_tokens && number_tag_position_list.size() > 1)
		fprintf(stderr, "warning: multiple number tags present in token name in \'%s\'\n", p_s_filename);
	// check presence of number tags in token name

	if(n_name_begin == n_name_end)
		return false;
	// unexpected characters

	while(b < e && isspace(r_s_token[b]))
		++ b;
	// skip space

	if(b < e && (r_s_token[b] == ':' || r_s_token[b] == '=')) {
		++ b;
		while(b < e && isspace(r_s_token[b]))
			++ b;
	}
	// skip a single ':', if present (present in GL_ARB_vertex_blend, maybe some fortran thing)
	// or a single '=', if present (present in GL_SGIS_pixel_texture)

	bool b_decimal = false;
	uint64_t n_value = 0;
	int n_length = 0;
	if(b + 2 < e && r_s_token[b] == '0' && r_s_token[b + 1] == 'x') {
		b += 2;
		// skip 0x

		if(b == e || !isxdigit(r_s_token[b]))
			return false;
		// no value?

		while(b < e && isxdigit(r_s_token[b])) {
			char n_char = tolower(r_s_token[b]);
			++ b;
			++ n_length;
			if(n_value > ((UINT64_MAX - 15) >> 4)) {
				fprintf(stderr, "error: too large hexadecimal number'\n");
				return false;
			}
			n_value <<= 4;
			n_value |= (isdigit(n_char))? n_char - '0' : n_char - 'a' + 10;
		}
		// parse value

		while(b < e && isspace(r_s_token[b]))
			++ b;
		// skip space
	} else if(b < e && isdigit(r_s_token[b])) {
		b_decimal = true;

		if(r_s_token[b] == '0') {
			while(b < e && isdigit(r_s_token[b])) {
				char n_char = r_s_token[b];
				if(n_char > '7') {
					fprintf(stderr, "error: corrupt base-8 number'\n");
					return false;
				}
				++ b;
				++ n_length;
				if(n_value > ((UINT64_MAX - 7) >> 3)) {
					fprintf(stderr, "error: too large octal number'\n");
					return false;
				}
				n_value <<= 3;
				n_value += n_char - '0';
			}
			// parse value (base 8 numers !!!)
		} else {
			while(b < e && isdigit(r_s_token[b])) {
				char n_char = r_s_token[b];
				++ b;
				++ n_length;
				if(n_value > (UINT64_MAX - 9) / 10) {
					fprintf(stderr, "error: too large decimal number'\n");
					return false;
				}
				n_value *= 10;
				n_value += n_char - '0';
			}
			// parse value (base 10 numbers)
		}
	} else {
		return false;
		// no value?
	}

	while(b < e && isspace(r_s_token[b]))
		++ b;
	// skip space

	if(b < e && r_s_token[b] == 'u') {
		if(b + 2 < e && r_s_token[b + 1] == 'l' && r_s_token[b + 1] == 'l')
			b += 3;
		else
			++ b;
		// skip 'u' / 'ull' suffix

		while(b < e && isspace(r_s_token[b]))
			++ b;
		// skip space
	}
	// skip 'u' / 'ull' suffix

	if(b_have_to_generate_multiple_tokens) {
		if(b < e && r_s_token[b] == '+') {
			++ b;
			// skip '+'

			while(b < e && isspace(r_s_token[b]))
				++ b;
			// skip space

			if(b + 3 <= e && r_s_token[b] == '<' && tolower(r_s_token[b + 1]) == 'n' && r_s_token[b + 2] == '>') {
				b_add_number_tag_to_value = true;
				b += 3;
			}
		}

		if(!b_add_number_tag_to_value)
			fprintf(stderr, "warning: number tag not present in token value in \'%s\'\n", p_s_filename);
	}
	// parse the number tag

	if(b + 1 < e && r_s_token[b] == '/' && r_s_token[b + 1] == '/')
		b = e;
	// "C" style line comment

	if(b != e)
		fprintf(stderr, "warning: comment afterwards token: %s\n", r_s_token.c_str());
	// that should be it (there may be comments)

	TToken t;
	t.b_decimal_value = b_decimal;
	t.n_value_digit_num = n_length;
	t.s_name.insert(t.s_name.begin(), r_s_token.begin() + n_name_begin,
		r_s_token.begin() + n_name_end);
	t.s_comment.insert(t.s_comment.begin(), r_s_token.begin() + b, r_s_token.begin() + e);
	if(t.s_name.find("GL_") == 0)
		t.s_name.erase(0, 3); // don't want GL_ there, we fill it in later
	t.n_value = n_value;
	if(b_have_to_generate_multiple_tokens) {
		fprintf(stderr, "warning: generating 32 versions of token \'%s\' in \'%s\'\n", t.s_name.c_str(), p_s_filename);
		// warn the user

		std::string s_pattern;
		s_pattern.swap(t.s_name);
		// get name pattern

		for(int i = 0; i < 32; ++ i) { // most of these are 32, but we can only guess
			char p_s_n[64];
			sprintf(p_s_n, "%d", i);
			// sptrintf <n>

			t.s_name = s_pattern;
			_ASSERTE(number_tag_position_list.size() <= INT_MAX);
			for(int i = int(number_tag_position_list.size()) - 1; i >= 0; -- i) {
				t.s_name.erase(number_tag_position_list[i], 3);
				// erase "<n>" occurences from the name (do it now so as to not disturb the position indexing)

				t.s_name.insert(number_tag_position_list[i], p_s_n);
				// insert the number instead
			}
			// generate a new name

			t.n_value = n_value + i;
			// generate a new value

			if(std::find_if(m_token_list.begin(), m_token_list.end(),
			   CFindTokenByName(t.s_name.c_str())) == m_token_list.end())
				m_token_list.push_back(t);
			// add it to the list; some tokens are detected multiple times, this handles that
		}
	} else {
		if(std::find_if(m_token_list.begin(), m_token_list.end(),
		   CFindTokenByName(t.s_name.c_str())) == m_token_list.end())
			m_token_list.push_back(t);
		// add it to the list; some tokens are detected multiple times, this handles that
	}
	// add token to the list

	return true;
}

bool CGLExtension::b_Have_ParamNames(const TFunction &r_t_func)
{
	for(size_t i = 0, n = r_t_func.param_list.size(); i < n; ++ i) {
		if(r_t_func.param_list[i].s_name.empty())
			return false;
	}

	return true;
}

bool CGLExtension::b_Have_Documentation(const TFunction &r_t_func)
{
	if(r_t_func.s_brief.empty())
		return false;
	for(size_t i = 0, n = r_t_func.param_list.size(); i < n; ++ i) {
		if(r_t_func.param_list[i].s_brief.empty())
			return false;
	}

	return true;
}

bool CGLExtension::b_Have_SomeDocumentation(const TFunction &r_t_func)
{
	if(!r_t_func.s_brief.empty())
		return true;
	for(size_t i = 0, n = r_t_func.param_list.size(); i < n; ++ i) {
		if(!r_t_func.param_list[i].s_brief.empty())
			return true;
	}

	return false;
}

bool CGLExtension::Make_DummyParamNames(TFunction &r_t_func)
{
	for(size_t i = 0, n = r_t_func.param_list.size(); i < n; ++ i) {
		if(r_t_func.param_list[i].s_name.empty()) {
			if(!stl_ut::Format(r_t_func.param_list[i].s_name, "__unnamed_%d", i))
				return false;
		}
	}

	return true;
}

const CXMLNode *CGLExtension::p_FindElemWithAttrib(const CXMLNode *p_parent, const char *p_s_node_name,
	const char *p_s_attrib_name, const char *p_s_attrib_value)
{
	for(const CXMLNode *p_elem = p_parent->p_FindChildElement(p_s_node_name);
	   CXMLNode::b_IsValidNode(p_elem); p_elem =
	   p_parent->p_FindNextChildElement(p_elem, p_s_node_name)) {
		const std::string *p_s_attrib = p_elem->p_Attrib(p_s_attrib_name);
		if(p_s_attrib && !p_s_attrib->compare(p_s_attrib_value))
			return p_elem;
	}

	return CXMLNode::p_InvalidNode();
}

bool CGLExtension::Retrieve_Doc_ParamNames(TFunction &r_t_func, bool b_need_doc, bool b_need_paranames,
	std::map<std::string, std::string> &r_fun_docu_url_map,
	const std::map<std::string, std::string> &r_fun_docu_alias_map) const // throws(std::bad_alloc)
{
	std::string s_xml_name;
	s_xml_name = "doc\\gl";
	s_xml_name += r_t_func.s_name;
	s_xml_name += ".xml";

	b_need_doc &= !b_Have_Documentation(r_t_func);
	b_need_paranames &= !b_Have_ParamNames(r_t_func);

	if(!b_need_doc && !b_need_paranames)
		return true;
	// another job well done

	/*if(!m_b_have_fun_docu_alias_map) {
		m_b_have_fun_docu_alias_map = true;
		m_fun_docu_alias_map["Uniform1d"] = "Uniform1f";
		m_fun_docu_alias_map["Uniform2d"] = "Uniform2f";
		m_fun_docu_alias_map["Uniform3d"] = "Uniform3f";
		m_fun_docu_alias_map["Uniform4d"] = "Uniform4f";
		m_fun_docu_alias_map["Uniform1dv"] = "Uniform1fv";
		m_fun_docu_alias_map["Uniform2dv"] = "Uniform2fv";
		m_fun_docu_alias_map["Uniform3dv"] = "Uniform3fv";
		m_fun_docu_alias_map["Uniform4dv"] = "Uniform4fv";
		m_fun_docu_alias_map["ProgramUniform1d"] = "ProgramUniform1f";
		m_fun_docu_alias_map["ProgramUniform2d"] = "ProgramUniform2f";
		m_fun_docu_alias_map["ProgramUniform3d"] = "ProgramUniform3f";
		m_fun_docu_alias_map["ProgramUniform4d"] = "ProgramUniform4f";
		m_fun_docu_alias_map["ProgramUniform1dv"] = "ProgramUniform1fv";
		m_fun_docu_alias_map["ProgramUniform2dv"] = "ProgramUniform2fv";
		m_fun_docu_alias_map["ProgramUniform3dv"] = "ProgramUniform3fv";
		m_fun_docu_alias_map["ProgramUniform4dv"] = "ProgramUniform4fv";
		m_fun_docu_alias_map["UniformMatrix2dv"] = "UniformMatrix2fv";
		m_fun_docu_alias_map["UniformMatrix3dv"] = "UniformMatrix3fv";
		m_fun_docu_alias_map["UniformMatrix4dv"] = "UniformMatrix4fv";
		m_fun_docu_alias_map["UniformMatrix2x3dv"] = "UniformMatrix2x3fv";
		m_fun_docu_alias_map["UniformMatrix3x2dv"] = "UniformMatrix3x2fv";
		m_fun_docu_alias_map["UniformMatrix2x4dv"] = "UniformMatrix2x4fv";
		m_fun_docu_alias_map["UniformMatrix4x2dv"] = "UniformMatrix4x2fv";
		m_fun_docu_alias_map["UniformMatrix3x4dv"] = "UniformMatrix3x4fv";
		m_fun_docu_alias_map["UniformMatrix4x3dv"] = "UniformMatrix4x3fv";
		m_fun_docu_alias_map["ProgramUniformMatrix2dv"] = "ProgramUniformMatrix2fv";
		m_fun_docu_alias_map["ProgramUniformMatrix3dv"] = "ProgramUniformMatrix3fv";
		m_fun_docu_alias_map["ProgramUniformMatrix4dv"] = "ProgramUniformMatrix4fv";
		m_fun_docu_alias_map["ProgramUniformMatrix2x3dv"] = "ProgramUniformMatrix2x3fv";
		m_fun_docu_alias_map["ProgramUniformMatrix3x2dv"] = "ProgramUniformMatrix3x2fv";
		m_fun_docu_alias_map["ProgramUniformMatrix2x4dv"] = "ProgramUniformMatrix2x4fv";
		m_fun_docu_alias_map["ProgramUniformMatrix4x2dv"] = "ProgramUniformMatrix4x2fv";
		m_fun_docu_alias_map["ProgramUniformMatrix3x4dv"] = "ProgramUniformMatrix3x4fv";
		m_fun_docu_alias_map["ProgramUniformMatrix4x3dv"] = "ProgramUniformMatrix4x3fv";
		m_fun_docu_alias_map["GetUniformdv"] = "GetUniformfv";
		m_fun_docu_alias_map["SamplerParameterIiv"] = "SamplerParameteriv";
		m_fun_docu_alias_map["SamplerParameterIuiv"] = "SamplerParameteriv";
		m_fun_docu_alias_map["GetSamplerParameterIiv"] = "GetSamplerParameteriv";
		m_fun_docu_alias_map["GetSamplerParameterIuiv"] = "GetSamplerParameteriv";
		m_fun_docu_alias_map["VertexAttribP1uiv"] = "VertexAttrib1sv";
		m_fun_docu_alias_map["VertexAttribP2uiv"] = "VertexAttrib2sv";
		m_fun_docu_alias_map["VertexAttribP3uiv"] = "VertexAttrib3sv";
		m_fun_docu_alias_map["VertexAttribP4uiv"] = "VertexAttrib4sv"; // *v versions yes, simple P versions no
	}*/
	// undocumented functions with very similar documented functions

	std::string s_wanted_function_name = p_s_Function_Name(r_t_func, false, true) + 2;
	if(r_fun_docu_alias_map.find(s_wanted_function_name) != r_fun_docu_alias_map.end())
		s_wanted_function_name = (*r_fun_docu_alias_map.find(s_wanted_function_name)).second; // can't use [] because of const
	// determine which function are we looking for

	bool b_downloaded_now = false;
	std::string s_url;
	if(!TFileInfo(s_xml_name.c_str()).b_exists) {
		/*if(!m_b_have_fun_docu_map) {
			m_b_have_fun_docu_map = true; // @t_odo - get rid of this hack
			m_fun_docu_map["GetBooleanv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml";
			m_fun_docu_map["GetBooleani_v"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml";
			m_fun_docu_map["GetDoublev"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml";
			m_fun_docu_map["GetFloatv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml";
			m_fun_docu_map["GetIntegerv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml";
			m_fun_docu_map["GetIntegeri_v"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml";
			//m_fun_docu_map["GetFloati_v"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml";
			//m_fun_docu_map["GetDoublei_v"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml";
			m_fun_docu_map["GetInteger64i_v"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml";
			m_fun_docu_map["GetInteger64v"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml";
			m_fun_docu_map["GetString"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml";

			m_fun_docu_map["UniformMatrix2fv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix3fv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix4fv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix2x3fv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix3x2fv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix2x4fv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix4x2fv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix3x4fv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix4x3fv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix2dv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix3dv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix4dv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix2x3dv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix3x2dv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix2x4dv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix4x2dv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix3x4dv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";
			m_fun_docu_map["UniformMatrix4x3dv"] = "http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml";

			m_fun_docu_map["ProgramUniformMatrix2fv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix3fv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix4fv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix2x3fv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix3x2fv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix2x4fv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix4x2fv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix3x4fv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix4x3fv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix2dv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix3dv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix4dv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix2x3dv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix3x2dv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix2x4dv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix4x2dv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix3x4dv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";
			m_fun_docu_map["ProgramUniformMatrix4x3dv"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml";

			m_fun_docu_map["VertexAttribIPointer"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glVertexAttribPointer.xml";
			m_fun_docu_map["VertexAttribLPointer"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glVertexAttribPointer.xml";

			m_fun_docu_map["Disable"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glEnable.xml";
			m_fun_docu_map["Disablei"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glEnable.xml";
			m_fun_docu_map["EndQuery"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glBeginQuery.xml";
			m_fun_docu_map["UnmapBuffer"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glMapBuffer.xml";
			m_fun_docu_map["ColorMaski"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glColorMask.xml";
			m_fun_docu_map["EndTransformFeedback"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glBeginTransformFeedback.xml";
			m_fun_docu_map["EndConditionalRender"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glBeginConditionalRender.xml";
			m_fun_docu_map["PauseTransformFeedback"] = "http://www.opengl.org/sdk/docs/man4/xhtml/glPauseTransformFeedback.xml";

			m_fun_docu_map["GetFramebufferAttachmentParameteriv"] = "http://www.khronos.org/opengles/sdk/docs/man/xhtml/glGetFramebufferAttachmentParameteriv.xml"; // GLES!
		}*/

		bool b_loop = true;

		for(int n_pass = 0; b_loop && n_pass < 2; ++ n_pass) {
			if(r_fun_docu_url_map.find(r_t_func.s_name.c_str()) != r_fun_docu_url_map.end()) {
				s_url = r_fun_docu_url_map[r_t_func.s_name];
				b_loop = false;
			} else {
				s_url = (n_pass)? "http://www.opengl.org/sdk/docs/man/xhtml/" :
					"http://www.opengl.org/sdk/docs/man4/xhtml/";
				s_url += p_s_Function_Name(r_t_func, false, true); // remove "ARB" from the name
				s_url += ".xml";
			}
			// get URL

			b_downloaded_now = true;

			printf("downloading \'%s\' ...\n", s_url.c_str());
			// verbose

			CDownloader downloader;
			int n_result;
			for(;;) {
				if(m_download_map.find(s_url) != m_download_map.end()) {
					const std::string &s_downloaded = m_download_map[s_url];
					printf("reusing \'%s\' as \'%s\'\n", s_downloaded.c_str(), s_xml_name.c_str());
					CopyFile(s_downloaded.c_str(), s_xml_name.c_str(), true);
					n_result = down_Success;
					break;
				}
				// reuse downloaded files (eg. glEnable and gldisable share the same documentation file)

				if(!downloader.StartDownload(s_xml_name.c_str(), s_url.c_str()) ||
				   (n_result = downloader.n_WaitForFinish()) == down_Failure) {
					fprintf(stderr, "error: failed to download \'%s\' (fatal error)\n", s_url.c_str());
					return false;
				}
				if(n_result == down_FileDoesntExist) {
					if(b_loop && s_url.length() > strlen("http://www.opengl.org/sdk/docs/man4/xhtml/gl") +
					   strlen(".xml") && strchr("1234IDNPLubsifdvx64", s_url[s_url.length() - 5])) {
						std::string s_old_url = s_url; // just for verbose
						s_url.erase(s_url.length() - 5, 1);
						fprintf(stderr, "warning: failed to download \'%s\': trying \'%s\'\n", s_old_url.c_str(), s_url.c_str());
						continue;
					}
					if(b_loop && !n_pass) {
						fprintf(stderr, "warning: failed to download \'%s\': trying to look elsewhere\n", s_url.c_str());
						break;
					}
					fprintf(stderr, "warning: failed to download \'%s\' (404)\n", s_url.c_str());
					return true;
				}
				_ASSERTE(n_result == down_Success);
				break;
			}
			if(n_result == down_Success) {
				m_download_map[s_url] = s_xml_name;
				break;
			}
		}
		// download the XML with function documentation
	}
	// make sure the XML does exist

	CXMLNode *p_doc;
	{
		CXML_EventParser parser(s_xml_name.c_str());
		if(!(p_doc = CXMLNode::p_Parse(parser)))
			return false;
	}
	// parse XML

	std::string s_func_brief;

	const CXMLNode *p_html = p_doc->p_FindChildElement("html");
	const CXMLNode *p_head = p_html->p_FindChildElement("head");

	const CXMLNode *p_redirect = p_FindElemWithAttrib(p_head, "meta", "http-equiv", "Refresh");
	if(!CXMLNode::b_IsValidNode(p_redirect))
		p_redirect = p_FindElemWithAttrib(p_head, "meta", "http-equiv", "refresh");
	const std::string *p_redirect_target;
	if(CXMLNode::b_IsValidNode(p_redirect) &&
	   (p_redirect_target = p_redirect->p_Attrib("content"))) {
		std::string s_where = *p_redirect_target;
		if(s_where.find(';') != std::string::npos)
			s_where.erase(0, s_where.find(';') + 1);
		if(s_where.find("=") != std::string::npos)
			s_where.erase(0, s_where.find('=') + 1);
		// get just an URL

		printf("warning: \'%s\' contains redirect to \'%s\' ...\n", s_url.c_str(), s_where.c_str());
		// verbose

		DeleteFile(s_xml_name.c_str());
		// delete the old file so it can be downloaded again

		if(s_url.rfind('/') != std::string::npos)
			s_url.erase(s_url.rfind('/'));
		r_fun_docu_url_map[r_t_func.s_name] = (s_url + "/" + s_where).c_str();
		// add record to redirect table

		return Retrieve_Doc_ParamNames(r_t_func, b_need_doc, b_need_paranames,
			r_fun_docu_url_map, r_fun_docu_alias_map);
		// recurse to retry
	}

	const CXMLNode *p_body = p_html->p_FindChildElement("body");
	const CXMLNode *p_refentry = p_FindElemWithAttrib(p_body, "div", "class", "refentry");
	for(size_t x = 0, y = p_refentry->n_Child_Num(); x < y; ++ x) {
		const CXMLNode *p_elem_x = p_refentry->p_Child(x);

		const std::string *p_s_class_x;
		if(!(p_elem_x->s_Name() == "div" && (p_s_class_x = p_elem_x->p_Attrib("class")) != 0 &&
		   (!p_s_class_x->compare("refsynopsisdiv") || !p_s_class_x->compare("refnamediv"))))
			continue;

		if(!p_s_class_x->compare("refnamediv")) {
			s_func_brief.clear();
			const CXMLNode *p_refnamediv = p_elem_x;//p_FindElemWithAttrib(p_refentry, "div", "class", "refnamediv");
			const CXMLNode *p_func_brief_para = p_refnamediv->p_FindChild("p");
			s_func_brief = p_func_brief_para->s_InnerText();
			DeUnicode(s_func_brief);
			TrimSpace(s_func_brief);
			// get brief

			int n_offset = 0, n_fb_len = s_func_brief.length();
			Skip_Identifier(n_offset, n_fb_len, s_func_brief);
			Skip_Space(n_offset, n_fb_len, s_func_brief);
			// skip the first function name

			while(n_offset < n_fb_len && s_func_brief[n_offset] == ',') {
				++ n_offset;
				Skip_Space(n_offset, n_fb_len, s_func_brief);
				if(n_offset + 1 < n_fb_len && s_func_brief[n_offset] == 'g' && s_func_brief[n_offset + 1] == 'l') {
					Skip_Identifier(n_offset, n_fb_len, s_func_brief);
					Skip_Space(n_offset, n_fb_len, s_func_brief);
				}
			}
			// handle multiple functions in a single brief

			if(n_offset < n_fb_len && s_func_brief[n_offset] == '-') { // long hyphen doesn't translate to '-'
				++ n_offset;
				Skip_Space(n_offset, n_fb_len, s_func_brief);
			}
			// handle the separator between the function name and the description

			s_func_brief.erase(0, n_offset);

			continue;
		}
		// get a brief description of what the function does

		const CXMLNode *p_refsynopsisdiv = p_elem_x;//p_FindElemWithAttrib(p_refentry, "div", "class", "refsynopsisdiv");

		for(size_t l = 0, p = p_refsynopsisdiv->n_Child_Num(); l < p; ++ l) { // more functions in a single file? (different way, found in http://www.opengl.org/sdk/docs/man4/xhtml/glTexParameter.xml)
			const CXMLNode *p_elem_l = p_refsynopsisdiv->p_Child(l);

			const std::string *p_s_class_l;
			if(!(p_elem_l->s_Name() == "div" && (p_s_class_l = p_elem_l->p_Attrib("class")) != 0 &&
			   !p_s_class_l->compare("funcsynopsis")))
				continue;

			const CXMLNode *p_funcsynopsis = p_elem_l;//p_FindElemWithAttrib(p_refsynopsisdiv, "div", "class", "funcsynopsis");

			for(size_t k = 0, o = p_funcsynopsis->n_Child_Num(); k < o; ++ k) { // more functions in a single file? (as in http://www.opengl.org/sdk/docs/man4/xhtml/glFramebufferTexture.xml)
				const CXMLNode *p_elem_k = p_funcsynopsis->p_Child(k);

				std::string s_function_name;
				// function name

				TDataType t_return_type;
				// return data type

				std::vector<std::pair<TDataType, std::string> > param_list;
				// list of function parameters

				bool b_valid = false;
				// is the parsed function valid?

				const std::string *p_s_class, *p_s_summary;
				if(p_elem_k->s_Name() == "table" && (p_s_class = p_elem_k->p_Attrib("class")) != 0 &&
				   !p_s_class->compare("funcprototype-table") || (p_s_summary =
				   p_elem_k->p_Attrib("summary")) != 0 && !p_s_summary->compare("Function synopsis")) {
					const CXMLNode *p_fprot_tab = p_elem_k;
					// get function prototype table

					bool b_parsed_all = true;
					for(size_t j = 0, m = p_fprot_tab->n_Child_Num(); j < m; ++ j) {
						const CXMLNode *p_tr = p_fprot_tab->p_ChildAt(j);
						if(p_tr->s_Name().compare("tr"))
							continue;
						// get table row

						_ASSERTE(p_tr->n_Child_Num() == 2 || p_tr->n_Child_Num() == 3);
						// left / right columns only, right?

						const CXMLNode *p_td_def = p_tr->p_ChildAt(0)->p_ToElement();
						const CXMLNode *p_td_type = p_tr->p_ChildAt(1)->p_ToElement();
						const CXMLNode *p_td_name = p_tr->p_ChildAt((p_tr->n_Child_Num() == 2)? 1 : 2)->p_ToElement();
						if(p_td_def->s_Name().compare("td") || p_td_type->s_Name().compare("td") ||
						   p_td_name->s_Name().compare("td")) {
							b_parsed_all = false;
							break;
						}
						// get table cells

						const CXMLNode *p_param = p_FindElemWithAttrib(p_td_name, "var", "class", "pdparam");
						if(!j) {
							const CXMLNode *p_funcdef = p_FindElemWithAttrib(p_td_def, "code", "class", "funcdef");
							const CXMLNode *p_funcname = p_FindElemWithAttrib(p_funcdef, "b", "class", "fsfunc");
							// parse the xml

							const CXMLNode *p_return_type_text = p_funcdef->p_FirstChild()->p_ToText();
							std::string s_return_type = p_return_type_text->s_Value();
							DeUnicode(s_return_type);
							TrimSpace(s_return_type);
							int b = 0, e = s_return_type.length();
							if(!e || !Skip_DataType(t_return_type, b, e, s_return_type) || b < e) {
								fprintf(stderr, "warning: failed to parse data type in XML doc: \'%s\'\n", s_return_type.c_str());
								b_parsed_all = false;
								break;
							}
							// parse data type the function returns

							s_function_name = p_funcname->s_InnerText();
							DeUnicode(s_function_name);
							// get function name
						}

						TDataType t_data_type;
						{
							const CXMLNode *p_datatype_text = CXMLNode::p_InvalidNode();
							if(p_tr->n_Child_Num() == 2) {
								for(size_t i = 0, n = p_td_name->n_Child_Num(); i < n; ++ i) {
									if(p_param == p_td_name->p_Child(i)) {
										if(i)
											p_datatype_text = p_td_name->p_Child(i - 1);
										break;
									}
								}
							}
							std::string s_data_type = (p_tr->n_Child_Num() == 3)? p_td_type->s_InnerText() : p_datatype_text->s_Value();
							if(s_data_type.empty())
								s_data_type = p_datatype_text->s_InnerText();
							DeUnicode(s_data_type);
							TrimSpace(s_data_type);
							int b = 0, e = s_data_type.length();
							if(!e || !Skip_DataType(t_data_type, b, e, s_data_type) || b < e) {
								fprintf(stderr, "warning: failed to parse data type in XML doc: \'%s\'\n", s_data_type.c_str());
								b_parsed_all = false;
								break;
							}
						}
						// get data type

						std::string s_param_name = p_param->s_InnerText();
						DeUnicode(s_param_name);
						TrimSpace(s_param_name);
						// get parameter name

						if(t_data_type.n_type != type_Void || t_data_type.n_pointer)
							param_list.push_back(std::make_pair(t_data_type, s_param_name));
						// add param names
					}
					if(b_parsed_all)
						b_valid = true;
					// functions with more than a single argument are in table
				} else if(p_elem_k->s_Name() == "p") {
					const CXMLNode *p_docu_para = p_elem_k;
					// get function prototype paragraph

					const CXMLNode *p_funcdef = p_FindElemWithAttrib(p_docu_para, "code", "class", "funcdef");
					const CXMLNode *p_funcname = p_FindElemWithAttrib(p_funcdef, "b", "class", "fsfunc");
					// parse the xml

					s_function_name = p_funcname->s_InnerText();
					DeUnicode(s_function_name);
					TrimSpace(s_function_name);
					// get function name

					bool b_parsed_all = true;
					for(const CXMLNode *p_param = p_docu_para->p_FindChildElement("var");
					   CXMLNode::b_IsValidNode(p_param); p_param =
					   p_docu_para->p_FindNextChildElement(p_param, "var")) {
						const std::string *p_s_param_class = p_param->p_Attrib("class");
						if(!p_s_param_class || p_s_param_class->compare("pdparam"))
							continue;
						// find parameter names

						if(!p_param->s_InnerText().compare("void")) {
							_ASSERTE(param_list.empty());
							break;
						}
						// handle void functions

						TDataType t_data_type;
						{
							const CXMLNode *p_datatype_text = CXMLNode::p_InvalidNode();
							for(size_t i = 0, n = p_docu_para->n_Child_Num(); i < n; ++ i) {
								if(p_param == p_docu_para->p_Child(i)) {
									if(i)
										p_datatype_text = p_docu_para->p_Child(i - 1);
									break;
								}
							}
							std::string s_data_type = p_datatype_text->s_Value();
							if(s_data_type.empty())
								s_data_type = p_datatype_text->s_InnerText();
							DeUnicode(s_data_type);
							TrimSpace(s_data_type);
							if(!s_data_type.empty() && s_data_type[0] == ',') {
								s_data_type.erase(0, 1);
								TrimSpace(s_data_type);
							}
							int b = 0, e = s_data_type.length();
							if(!e || !Skip_DataType(t_data_type, b, e, s_data_type) || b < e) {
								fprintf(stderr, "warning: failed to parse data type in XML doc: \'%s\'\n", s_data_type.c_str());
								b_parsed_all = false;
								break;
							}
						}
						// get data type

						std::string s_param_name = p_param->s_InnerText();
						DeUnicode(s_param_name);
						TrimSpace(s_param_name);
						// get parameter name

						if(t_data_type.n_type != type_Void || t_data_type.n_pointer)
							param_list.push_back(std::make_pair(t_data_type, s_param_name));
						// add param names
					}
					// colect function parameters

					const CXMLNode *p_return_type_text = p_funcdef->p_FirstChild()->p_ToText();
					std::string s_return_type = p_return_type_text->s_Value();
					DeUnicode(s_return_type);
					TrimSpace(s_return_type);
					int b = 0, e = s_return_type.length();
					if(!e || !Skip_DataType(t_return_type, b, e, s_return_type) || b < e) {
						fprintf(stderr, "warning: failed to parse data type in XML doc: \'%s\'\n", s_return_type.c_str());
						b_parsed_all = false;
					}
					// parse data type the function returns

					if(b_parsed_all)
						b_valid = true;
				} else
					continue; // other element
				// parse function

				if(!b_valid) {
					fprintf(stderr, "warning: \'%s\': function declaration parser failed\n", s_xml_name.c_str());
					continue;
				}
				// was it parsed correctly?

				if(b_downloaded_now) {
					printf("found %s%s(", p_s_DataType_Name_Space(t_return_type), s_function_name.c_str());
					for(size_t i = 0, n = param_list.size(); i < n; ++ i) {
						printf("%s%s%s", (i)? ", " : "", p_s_DataType_Name_Space(param_list[i].first),
							param_list[i].second.c_str());
					}
					printf(")\n");
				}
				// verbose

				if(s_function_name.length() >= 2 && s_function_name[0] == 'g' && s_function_name[1] == 'l')
					s_function_name.erase(0, 2);
				// erase 'gl' prefix if present

				if(//t_return_type == r_t_func.t_return_type && // types do not compare now (todo)
				   s_function_name == s_wanted_function_name &&
				   param_list.size() == r_t_func.param_list.size()) {
					// @todo - compare parameter types!

					if(b_need_paranames) {
						for(size_t i = 0, n = param_list.size(); i < n; ++ i)
							r_t_func.param_list[i].s_name = param_list[i].second;
						// copy parameter names
					}

					if(b_need_doc) {
						std::map<std::string, std::string> doc_list;

						r_t_func.s_brief = s_func_brief;
						// @t_odo - get function brief from refnamediv (use the long hyphen character to skip over the function name)

						bool b_found_refsect1 = false;
						for(size_t z = x + 1; z < y; ++ z) {
							const CXMLNode *p_elem_z = p_refentry->p_Child(z);

							const std::string *p_s_class_z;
							if(!(p_elem_z->s_Name() == "div" && (p_s_class_z = p_elem_z->p_Attrib("class")) != 0 &&
							   !p_s_class_z->compare("refsect1")))
								continue;
							const CXMLNode *p_refsect1 = p_elem_z;
							// get a refsect1 div

							const CXMLNode *p_a = p_refsect1->p_FindChildElement("a");
							const std::string *p_s_id;
							if((!(p_s_id = p_a->p_Attrib("id")) && !(p_s_id = p_a->p_Attrib("name"))) ||
							   p_s_id->find("parameters") != 0) // they use parameters, parameters2, parameters3, ...
								continue;
							// make sure it is paragraph with parameters documentation

							b_found_refsect1 = true;
							const CXMLNode *p_varlist = p_FindElemWithAttrib(p_refsect1, "div", "class", "variablelist");
							const CXMLNode *p_dl = p_varlist->p_FindChildElement("dl");
							_ASSERTE(!(p_dl->n_Child_Num() % 2)); // should be even - just pairs of <dt> and <dd> (note the code below won't crash if it's not)
							for(size_t v = 0, w = p_dl->n_Child_Num(); v < w; v += 2) {
								const CXMLNode *p_dt = p_dl->p_Child(v)->p_ToElement();
								const CXMLNode *p_dd = p_dl->p_ChildAt(v + 1)->p_ToElement();
								if(p_dt->s_Name().compare("dt") || p_dd->s_Name().compare("dd")) {
									-- v; // try to reync
									fprintf(stderr, "warning: \'%s\': unsynced data list\n", s_xml_name.c_str());
									continue;
								}

								std::string s_pname = s_InnerText(p_dt);
								DeUnicode(s_pname);
								TrimSpace(s_pname);

								{
									size_t n_position = 0;
									while((n_position = s_pname.find(' ', n_position)) != std::string::npos)
										s_pname.erase(n_position, 1);
								}
								// remove whitespace

								std::string s_brief = s_InnerText(p_dd);
								DeUnicode(s_brief);
								TrimSpace(s_brief);

								{
									size_t n_position = 0;
									while((n_position = s_brief.find('\n', n_position)) != std::string::npos)
										s_brief[n_position] = ' ';
									n_position = 0;
									while((n_position = s_brief.find('\r', n_position)) != std::string::npos)
										s_brief[n_position] = ' ';
									n_position = 0;
									while((n_position = s_brief.find('\t', n_position)) != std::string::npos)
										s_brief[n_position] = ' ';
									n_position = 0;
									while((n_position = s_brief.find((unsigned char)0xa0, n_position)) != std::string::npos) // nbsp
										s_brief[n_position] = ' ';
									n_position = 0;
									while((n_position = s_brief.find("  ", n_position)) != std::string::npos)
										s_brief.erase(n_position, 1);
								}
								// condense whitespace

								if(s_pname.find(',') != std::string::npos) {
									// a list of multiple parameters

									std::vector<std::string> arg_name_list;
									if(!stl_ut::Split(arg_name_list, s_pname, ",", 0))
										throw std::bad_alloc();
									// split string using a separator

									std::for_each(arg_name_list.begin(), arg_name_list.end(), TrimSpace); // did it already for spaces, there might be more
									// remove whitespace

									for(size_t i = 0, n = arg_name_list.size(); i < n; ++ i) {
										if(arg_name_list[i].empty())
											continue;
										doc_list[arg_name_list[i]] = s_brief;
									}
									// add the same brief for all the params
								} else
									doc_list[s_pname] = s_brief;
							}

							break;
						}
						if(!b_found_refsect1 && !r_t_func.param_list.empty())
							fprintf(stderr, "warning: \'%s\': doesn't contain parameters documentation\n", s_xml_name.c_str());
						// @t_odo - parse documentation (it is in a refsect1 div after this refsynopsis div)

						// @t_odo - think about return code // the return code is poorly documented in OpenGL man pages

						bool b_missing_params = false;
						for(size_t i = 0, n = param_list.size(); i < n; ++ i) {
							if(doc_list.find(r_t_func.param_list[i].s_name) == doc_list.end()) {
								b_missing_params = true;
								continue;
							}
							// don't copy empty params

							r_t_func.param_list[i].s_brief = doc_list[r_t_func.param_list[i].s_name];
						}
						// copy parameter names

						if(b_downloaded_now || b_missing_params) {
							printf("in \'%s\':\n", m_s_ext_name.c_str());
							PrintFunctionDoc(r_t_func, stdout, p_s_Function_Name(r_t_func, false, false));
							printf("%s%s(", p_s_DataType_Name_Space(t_return_type), s_function_name.c_str());
							for(size_t i = 0, n = param_list.size(); i < n; ++ i) {
								printf("%s%s%s", (i)? ", " : "", p_s_DataType_Name_Space(param_list[i].first),
									param_list[i].second.c_str());
							}
							printf(")\n");
						}
						// dump documentation
					}

					delete p_doc;
					// cleanup;

					return true;
				}
				// in case the correct function was found, 
			}
		}
	}

	if(b_need_doc && b_need_paranames)
		fprintf(stderr, "warning: arg names and documentation not found for %s()\n", r_t_func.s_name.c_str());
	else if(b_need_doc)
		fprintf(stderr, "warning: documentation not found for %s()\n", r_t_func.s_name.c_str());
	else /*if(b_need_paranames)*/ {
		_ASSERTE(b_need_paranames);
		fprintf(stderr, "warning: arg names not found for %s()\n", r_t_func.s_name.c_str());
	}

	delete p_doc;
	// fail

	return false;
}

void CGLExtension::PrintFunctionDoc(const TFunction &r_t_func, FILE *p_fw, const char *p_s_token_name)
{
	if(p_fw != stdout)
		fprintf(p_fw, "\n");

	fprintf(p_fw, "/**\n");
	if(p_s_token_name)
		fprintf(p_fw, " *  @def %s\n", p_s_token_name);
	fprintf(p_fw, " *  @brief %s\n", r_t_func.s_brief.c_str());
	for(size_t i = 0, n = r_t_func.param_list.size(); i < n; ++ i) {
		int n_position = fprintf(p_fw, " *  @param[%s] %s ",
			(r_t_func.param_list[i].t_data_type.n_pointer &&
			!r_t_func.param_list[i].t_data_type.b_const)? "out" : "in",
			r_t_func.param_list[i].s_name.c_str()) /*+ 1*/; // +1 corrects for the tab character which is counted as a single one // no longer tabs
		if(r_t_func.param_list[i].s_brief.empty()) {
			fprintf(p_fw, "(documantation not found)\n", r_t_func.param_list[i].s_name.c_str());
			continue;
		}
		const char *p_s_doc = r_t_func.param_list[i].s_brief.c_str();
		const int n_keep_under = 100;
		while(*p_s_doc) {
			int n_remaining = n_keep_under - n_position;
			const char *p_s_break = strchr(p_s_doc, ' ');
			if(!p_s_break) {
				fprintf(p_fw, "%s\n", p_s_doc);
				break;
			}
			if(p_s_break - p_s_doc < n_remaining) {
				const char *p_s_prev_break = p_s_break;
				for(;;) {
					p_s_break = strchr(p_s_break + 1, ' ');
					if(!p_s_break) {
						p_s_break = p_s_prev_break + strlen(p_s_prev_break); // print it to the end
						break;
					}
					if(p_s_break - p_s_doc > n_remaining) {
						p_s_break = p_s_prev_break; // break the line
						break;
					}
					p_s_prev_break = p_s_break;
				}
			}
			std::string s_print;
			s_print.insert(s_print.begin(), p_s_doc, p_s_break);
			fprintf(p_fw, "%s\n", s_print.c_str());
			p_s_doc = (*p_s_break == ' ')? p_s_break + 1 : p_s_break; // skip the leading space
			if(*p_s_doc) {
				n_position = fprintf(p_fw, " *      "); // no longer tabs
				//n_position = 8; // tab characters are miscounted
			}
		}
	}
	fprintf(p_fw, " */\n");
}

inline bool CGLExtension::isspace_or_separator(char c)
{
	return c == 0xa0 || isspace(c) || c == '.' || c == ';' || c == ',';
}

std::string CGLExtension::s_InnerText(const CXMLNode *p_node) // throws(std::bad_alloc)
{
	if(p_node->b_IsText())
		return p_node->s_Value();
	if(p_node->b_IsElem()) {
		if(!p_node->n_Child_Num())
			return p_node->s_InnerText();
		std::string s_text;
		for(size_t i = 0, n = p_node->n_Child_Num(); i < n; ++ i) {
			const CXMLNode *p_child = p_node->p_Child(i);
			const std::string &s_child_text = s_InnerText(p_child);
			if(!s_text.empty() && !s_child_text.empty() &&
			   !isspace_or_separator(s_text[s_text.length() - 1]) &&
			   !isspace_or_separator(s_child_text[s_child_text.length() - 1]))
				s_text += ' ';
			s_text += s_child_text;
		}
		return s_text;
	} else
		throw std::runtime_error("s_InnerText failure");
}

bool CGLExtension::Parse_Procedure(const std::string &_r_s_proc_,
	const char *p_s_filename, const std::set<std::string> &r_deprecated_func_set,
	bool b_find_documentation, std::map<std::string, std::string> &r_fun_docu_url_map,
	const std::map<std::string, std::string> &r_fun_docu_alias_map) // throws(std::bad_alloc)
{
	//m_s_proc_string += _r_s_proc_;
	const std::string &r_s_proc = _r_s_proc_;//m_s_proc_string;
	int b = 0, e = r_s_proc.length();
	// procedure definition may span multiple lines

	Skip_Space(b, e, r_s_proc);
	if(b == e) {
		m_b_proc_complete = false;
		return true;
	}
	// incomplete definition

	if(m_n_proc_completeness == 0) {
		if(!Skip_DataType(m_t_parsed_function.t_return_type, b, e, r_s_proc)) {
			fprintf(stderr, "error: failed to parse data type in \'%s\'\n", p_s_filename);
			return false;
		}
		// get return data type

		m_n_proc_completeness = 1;
	}

	if(m_n_proc_completeness == 1) {
		Skip_Space(b, e, r_s_proc);
		if(b == e) {
			m_b_proc_complete = false;
			return true;
		}
		// incomplete definition

		int n_name_begin = b;
		if(!Skip_Identifier(b, e, r_s_proc)) {
			fprintf(stderr, "error: failed to parse function name in \'%s\'\n", p_s_filename);
			return false;
		}
		if(n_name_begin + 2 < b && r_s_proc[n_name_begin] == 'g' && r_s_proc[n_name_begin + 1] == 'l')
			n_name_begin += 2; // skip 'gl' from the function name; will preppend it later
		m_t_parsed_function.s_name.erase();
		m_t_parsed_function.s_name.insert(m_t_parsed_function.s_name.begin(),
			r_s_proc.begin() + n_name_begin, r_s_proc.begin() + b);
		// get function name

		Skip_Space(b, e, r_s_proc);
		if(b < e && r_s_proc[b] == '{' || r_s_proc[b] == '[') {
			char n_closing_char = (r_s_proc[b] == '{')? '}' : ']';
			++ b;

			for(;;) {
				Skip_Space(b, e, r_s_proc);
				if(b == e) {
					fprintf(stderr, "error: unexpected end before \'%c\' in \'%s\'\n", n_closing_char);
					return false;
				}
				if(r_s_proc[b] == n_closing_char) {
					++ b;
					break;
				} else if(n_closing_char == ']' && r_s_proc[b] == '|') { // ATI occasionaly uses regex-like [ub|s|i|f|d] format
					++ b;
					continue;
				}
				// look for the end

				EDataType n_type;
				if(tolower(r_s_proc[b]) == 'u') {
					++ b;
					if(b == e) {
						fprintf(stderr, "error: unexpected end before \'%c\' in \'%s\'\n", n_closing_char, p_s_filename);
						return false;
					}

					switch(tolower(r_s_proc[b])) {
					case 'b':
						n_type = type_UByte;
						break;
					case 's':
						n_type = type_UShort;
						break;
					case 'i':
						n_type = type_UInt;
						break;
					default:
						fprintf(stderr, "error: unknown data type \'u%c\' in \'%s\'\n", r_s_proc[b], p_s_filename);
						return false;
					}
					++ b;
				} else {
					switch(r_s_proc[b]) {
					case 'b':
						n_type = type_Byte;
						break;
					case 's':
						n_type = type_Short;
						break;
					case 'i':
						n_type = type_Int;
						break;
					case 'f':
						n_type = type_Float;
						break;
					case 'd':
						n_type = type_Double;
						break;
					default:
						if(isdigit(r_s_proc[b])) {
							fprintf(stderr, "error: vector template function generator not "
								"implemented; please edit %s in \'%s\' manually\n", r_s_proc[b], p_s_filename);
							// glVertex{1234}{ubsifdx}[v](T coords)
							// this is hard to parse correctly, and even harder to generate
							// good names for the arguments
							// @todo - implement it
							return false;
						}
						fprintf(stderr, "error: unknown data type \'%c\' in \'%s\'\n", r_s_proc[b], p_s_filename);
						return false;
					}
					++ b;
				}
				// recognize data type

				m_parsed_function_variant_list.push_back(n_type);
			}

			if(b < e && !isspace(r_s_proc[b]) && r_s_proc[b] != '(') {
				int n_name_begin = b;
				if(!Skip_Identifier(b, e, r_s_proc)) {
					fprintf(stderr, "error: failed to parse function name (2) in \'%s\'\n", p_s_filename);
					return false;
				}
				m_parsed_function_name_append.erase();
				m_parsed_function_name_append.insert(m_parsed_function_name_append.begin(),
					r_s_proc.begin() + n_name_begin, r_s_proc.begin() + b);
			} else
				m_parsed_function_name_append.erase();
			// parse name value to append ("ARB" for glSecondaryColor3[fdb]ARB)
		}
		// parse variant list, such as glSecondaryColor3[fdb]ARB

		m_n_proc_completeness = 2;
	}

	if(m_n_proc_completeness == 2) {
		Skip_Space(b, e, r_s_proc);
		if(b == e) {
			m_b_proc_complete = false;
			return true;
		}
		// incomplete definition

		if(r_s_proc[b] != '(') {
			fprintf(stderr, "error: unexpected character(s) found before \'(\' in \'%s\'\n", p_s_filename);
			return false;
		}
		++ b;
		// skip opening parenthesis

		m_n_proc_completeness = 3;
	}

	if(m_n_proc_completeness < 6) for(;;) {
		Skip_Space(b, e, r_s_proc);
		if(b == e) {
			m_b_proc_complete = false;
			return true;
		}
		// incomplete definition

		if(m_n_proc_completeness == 3) {
			if(r_s_proc[b] == ')') {
				++ b;
				m_n_proc_completeness = 6;
				break;
			}
			// closing parenthesis

			if(!Skip_DataType(m_t_parsed_param_data_type, b, e, r_s_proc)) {
				fprintf(stderr, "error: failed to parse function parameter data type in \'%s\'\n", p_s_filename);
				return false;
			}
			// parse data type

			// there is a problem right here; data type may span multiple lines,
			// user needs to occasionaly edit some extension files for
			// them to parse properly (hovewer, it always reports error of
			// some kind (unexpected identifier / data type / token))
			// @todo - rewrite the parser to always read the whole segments
			//		between '(' or ',' and the next ',' or ')' and parse them then

			m_n_proc_completeness = 4;
		}
		// skip parameter data type

		Skip_Space(b, e, r_s_proc);
		if(b == e) {
			m_b_proc_complete = false;
			return true;
		}
		// incomplete definition

		do {
			if(m_n_proc_completeness == 4) {
				if(r_s_proc[b] == ')') {
					++ b;
					m_n_proc_completeness = 6;

					TParam p;
					p.t_data_type = m_t_parsed_param_data_type;
					if(p.t_data_type.n_type != type_Void || p.t_data_type.n_pointer)
						m_t_parsed_function.param_list.push_back(p);
					// have a param without name
					break;
				}
				// closing parenthesis

				if(r_s_proc[b] == ',') {
					++ b;
					m_n_proc_completeness = 3;

					TParam p;
					p.t_data_type = m_t_parsed_param_data_type;
					if(p.t_data_type.n_type != type_Void || p.t_data_type.n_pointer)
						m_t_parsed_function.param_list.push_back(p);
					// have a param without name

					break;
				}
				// next parameter

				int n_name_begin = b;
				if(!Skip_Identifier(b, e, r_s_proc)) {
					fprintf(stderr, "error: failed to parse function parameter name in \'%s\'\n", p_s_filename);
					return false;
				}
				TParam p;
				p.t_data_type = m_t_parsed_param_data_type;
				p.s_name.erase();
				p.s_name.insert(p.s_name.begin(),
					r_s_proc.begin() + n_name_begin, r_s_proc.begin() + b);
				// get parameter name

				Skip_Space(b, e, r_s_proc);
				while(r_s_proc[b] == '[') {
					++ b;
					while(b < e && isdigit(r_s_proc[b]))
						++ b;
					if(b == e) {
						fprintf(stderr, "error: unexpected end before \']' in \'%s\'\n", p_s_filename);
						return false;
					}
					if(r_s_proc[b] != ']') {
						fprintf(stderr, "error: unexpected character(s) found before \']' in \'%s\'\n", p_s_filename);
						return false;
					}
					++ b;

					++ p.t_data_type.n_pointer; // represent it as pointer in "C"
				}
				// there might be [] or [16] after the name

				if(p.t_data_type.n_type != type_Void || p.t_data_type.n_pointer)
					m_t_parsed_function.param_list.push_back(p);
				// add it to the list

				m_n_proc_completeness = 5;

			}
			// skip parameter name
		} while(0);

		if(m_n_proc_completeness == 5) {
			Skip_Space(b, e, r_s_proc);
			if(b == e) {
				m_b_proc_complete = false;
				return true;
			}
			// incomplete definition

			if(r_s_proc[b] == ')') {
				++ b;
				m_n_proc_completeness = 6;
				break;
			}
			// closing parenthesis

			if(r_s_proc[b] != ',') {
				fprintf(stderr, "error: unexpected character's found before \',\' in \'%s\'\n", p_s_filename);
				return false;
			}
			++ b;
			m_n_proc_completeness = 3;
			// next parameter
		}

		if(m_n_proc_completeness == 6)
			break;
	}
	// parse function parameters until ')'

	bool b_deprecated = r_deprecated_func_set.find(m_t_parsed_function.s_name) !=
		r_deprecated_func_set.end();

	if((!b_Have_ParamNames(m_t_parsed_function) || !b_Have_Documentation(m_t_parsed_function)) &&
	   !Retrieve_Doc_ParamNames(m_t_parsed_function, !b_deprecated && b_find_documentation,
	   !b_Have_ParamNames(m_t_parsed_function), r_fun_docu_url_map, r_fun_docu_alias_map) &&
	   (!b_Have_ParamNames(m_t_parsed_function) && !Make_DummyParamNames(m_t_parsed_function))) {
		fprintf(stderr, "error: failed to retrieve function parameter names\n");
		return false;
	}
	// make sure the parameters do have names, find documentation (not for deprecated functions)

	Skip_Space(b, e, r_s_proc);
	if(b < e && r_s_proc[b] == ';')
		++ b;
	Skip_Space(b, e, r_s_proc);
	// skip semicolon at the end

	if(b != e) {
		fprintf(stderr, "warning: suspicious values afterwards "
			"function definition: %s in \'%s\'\n", r_s_proc.c_str(), p_s_filename);
	}
	// that should be it (there may be comments)

	if(!m_parsed_function_variant_list.empty()) {
		for(size_t i = 0, n = m_parsed_function_variant_list.size(); i < n; ++ i) {
			EDataType n_template_data_type = m_parsed_function_variant_list[i];

			TFunction f = m_t_parsed_function;
			// make a copy

			switch(n_template_data_type) {
			case type_UByte:
				f.s_name += "ub";
				break;
			case type_UShort:
				f.s_name += "us";
				break;
			case type_UInt:
				f.s_name += "ui";
				break;
			case type_Byte:
				f.s_name += 'b';
				break;
			case type_Short:
				f.s_name += 's';
				break;
			case type_Int:
				f.s_name += 'i';
				break;
			case type_Float:
				f.s_name += 'f';
				break;
			case type_Double:
				f.s_name += 'd';
				break;
			}
			f.s_name += m_parsed_function_name_append;
			// create name

			for(size_t j = 0, m = f.param_list.size(); j < m; ++ j) {
				if(f.param_list[j].t_data_type.n_type == type_TemplateType)
					f.param_list[j].t_data_type.n_type = n_template_data_type;
			}
			// exchange template types for non-template
			
			m_function_list.push_back(f);
			// put this variant to the list
		}
	} else {
#ifdef _DEBUG
		for(size_t j = 0, m = m_t_parsed_function.param_list.size(); j < m; ++ j)
			_ASSERTE(m_t_parsed_function.param_list[j].t_data_type.n_type != type_TemplateType);
#endif //_DEBUG
		m_function_list.push_back(m_t_parsed_function);
	}
	// add function to the list

	Reset_FunctionParser();
	// !!

	return true;
}

void CGLExtension::Reset_FunctionParser()
{
	m_parsed_function_variant_list.clear();
	m_parsed_function_name_append.erase();
	m_s_proc_string.erase();
	m_b_proc_complete = true;
	m_n_proc_completeness = 0;
	m_t_parsed_function.t_return_type.n_type = (EDataType)-1;
	m_t_parsed_function.s_name.erase();
	m_t_parsed_function.s_brief.erase();
	m_t_parsed_function.param_list.clear();
	m_t_parsed_param_data_type.n_type = (EDataType)-1;
	// reset function parser state
}

bool CGLExtension::b_FunctionParser_NeedsMore() const
{
	return !m_b_proc_complete;
}

void CGLExtension::Skip_Space(int &b, int e, const std::string &r_s_proc)
{
	while(b < e && isspace(r_s_proc[b]))
		++ b;
}

bool CGLExtension::Skip_Identifier(int &b, int e, const std::string &r_s_proc)
{
	Skip_Space(b, e, r_s_proc);
	if(b == e)
		return false; // unexpected end
	if(!isalpha(r_s_proc[b]) && r_s_proc[b] != '_')
		return false; // bad first char
	++ b;
	while(b < e && (isalpha(r_s_proc[b]) || isdigit(r_s_proc[b]) || r_s_proc[b] == '_'))
		++ b;
	return true;
}

const char *CGLExtension::p_s_Token_Name(const TToken &r_token,
	bool b_force_brand, bool b_remove_brand) const // throws(std::bad_alloc)
{
	_ASSERTE(!(b_force_brand && b_remove_brand));
	_ASSERTE(r_token.s_name.size() < 3 || r_token.s_name[0] != 'G' ||
		r_token.s_name[1] != 'L' || r_token.s_name[2] != '_'); // shouldn't start with "GL_"

	if(b_force_brand || b_remove_brand) {
		bool b_have_brand = false;
		if(r_token.s_name.length() > m_s_ext_brand.length()) {
			size_t n_assumed_brand_pos = r_token.s_name.length() - m_s_ext_brand.length();
			b_have_brand = r_token.s_name[n_assumed_brand_pos - 1] == '_' &&
				r_token.s_name.find(m_s_ext_brand, n_assumed_brand_pos) == n_assumed_brand_pos;
		}
		// determine whether the function has a brand

		if((!b_have_brand && b_remove_brand) ||
		   (b_have_brand && b_force_brand)) {
			if(!stl_ut::Format(m_s_temp_tname_string, "GL_%s", r_token.s_name.c_str()))
				throw std::bad_alloc();
			return m_s_temp_tname_string.c_str();
		}
		// handle the easy scenarios when the function already is in the desired form

		if(!b_have_brand && b_force_brand) {
			if(!stl_ut::Format(m_s_temp_tname_string, "GL_%s_%s",
			   r_token.s_name.c_str(), m_s_ext_brand.c_str()))
				throw std::bad_alloc();
			return m_s_temp_tname_string.c_str();
		}
		// just add the brand

		_ASSERTE(b_have_brand && b_remove_brand); // there's no other option, right?
		/*if(b_have_brand && b_remove_brand)*/ {
			m_s_temp_tname_string = "GL_";
			m_s_temp_tname_string.insert(m_s_temp_tname_string.end(),
				r_token.s_name.begin(), r_token.s_name.begin() +
				r_token.s_name.length() - (m_s_ext_brand.length() + 1));
			return m_s_temp_tname_string.c_str();
		}
		// remove the brand
	} else {
		stl_ut::Format(m_s_temp_tname_string, "GL_%s", r_token.s_name.c_str());
		return m_s_temp_tname_string.c_str();
	}
}

const char *CGLExtension::p_s_Function_Name(const TFunction &r_func,
	bool b_force_brand, bool b_remove_brand) const // throws(std::bad_alloc)
{
	_ASSERTE(!(b_force_brand && b_remove_brand));
	_ASSERTE(r_func.s_name.size() < 2 || r_func.s_name[0] != 'g' ||
		r_func.s_name[1] != 'l'); // shouldn't start with "gl"

	if(b_force_brand || b_remove_brand) {
		bool b_have_brand = false;
		if(r_func.s_name.length() > m_s_ext_brand.length()) {
			size_t n_assumed_brand_pos = r_func.s_name.length() - m_s_ext_brand.length();
			b_have_brand = r_func.s_name.find(m_s_ext_brand,
				n_assumed_brand_pos) == n_assumed_brand_pos;
		}
		// determine whether the function has a brand

		if((!b_have_brand && b_remove_brand) ||
		   (b_have_brand && b_force_brand)) {
			if(!stl_ut::Format(m_s_temp_fname_string, "gl%s", r_func.s_name.c_str()))
				throw std::bad_alloc();
			return m_s_temp_fname_string.c_str();
		}
		// handle the easy scenarios when the function already is in the desired form

		if(!b_have_brand && b_force_brand) {
			if(!stl_ut::Format(m_s_temp_fname_string, "gl%s%s",
			   r_func.s_name.c_str(), m_s_ext_brand.c_str()))
				throw std::bad_alloc();
			return m_s_temp_fname_string.c_str();
		}
		// just add the brand

		_ASSERTE(b_have_brand && b_remove_brand); // there's no other option, right?
		/*if(b_have_brand && b_remove_brand)*/ {
			m_s_temp_fname_string = "gl";
			m_s_temp_fname_string.insert(m_s_temp_fname_string.end(),
				r_func.s_name.begin(), r_func.s_name.begin() +
				r_func.s_name.length() - m_s_ext_brand.length());
			return m_s_temp_fname_string.c_str();
		}
		// remove the brand
	} else {
		stl_ut::Format(m_s_temp_fname_string, "gl%s", r_func.s_name.c_str());
		return m_s_temp_fname_string.c_str();
	}
}

const char *CGLExtension::p_s_DataType_Name(TDataType t_type) const
{
	std::string &r_s_name = m_s_temp_data_type_string;
	// just to make it shorter

	for(size_t i = 0; i < sizeof(p_type_name_table) / sizeof(p_type_name_table[0]); ++ i) {
		if(p_type_name_table[i].n_type == t_type.n_type) {
			/*if(!t_type.b_const && !t_type.n_pointer)
				return p_type_name_table[i].p_s_name;*/

			r_s_name.erase(); // clear it !!
			if(t_type.b_const)
				r_s_name = "const ";
			if((t_type.n_type < type_FirstNativeType || t_type.n_type > type_LastNativeType) &&
			   t_type.n_type != type_Void && t_type.n_type != type_Int && t_type.n_type != type_Short &&
			   t_type.n_type != type_Float && t_type.n_type != type_Double)
				r_s_name += "GL"; // it's one of GL types (like GLenum)
			r_s_name += p_type_name_table[i].p_s_name;
			for(int j = 0; j < t_type.n_pointer; ++ j)
				r_s_name += "*";
			// concatenate here (unsafe - todo :: use member std::string)

			return r_s_name.c_str();
		}
	}

	return "<unknown type>";
}

const char *CGLExtension::p_s_DataType_Name_Space(TDataType t_type) const
{
	std::string &r_s_name = m_s_temp_data_type_string;
	// just to make it shorter

	for(size_t i = 0; i < sizeof(p_type_name_table) / sizeof(p_type_name_table[0]); ++ i) {
		if(p_type_name_table[i].n_type == t_type.n_type) {
			/*if(!t_type.b_const && !t_type.n_pointer)
				return p_type_name_table[i].p_s_name;*/

			r_s_name.erase(); // clear it !!
			if(t_type.b_const)
				r_s_name = "const ";
			if((t_type.n_type < type_FirstNativeType || t_type.n_type > type_LastNativeType) &&
			   t_type.n_type != type_Void && t_type.n_type != type_Int && t_type.n_type != type_Short &&
			   t_type.n_type != type_Float && t_type.n_type != type_Double)
				r_s_name += "GL"; // it's one of GL types (like GLenum)
			r_s_name += p_type_name_table[i].p_s_name;
			if(t_type.n_pointer)
				r_s_name += " ";
			for(int j = 0; j < t_type.n_pointer; ++ j)
				r_s_name += "*";
			if(!t_type.n_pointer)
				r_s_name += " ";
			// concatenate here (unsafe - todo :: use member std::string)

			return r_s_name.c_str();
		}
	}

	return "<unknown type>";
}

bool CGLExtension::Skip_DataType(TDataType &r_t_type, int &b, int e, const std::string &r_s_proc)
{
	r_t_type.b_const = false;
	r_t_type.n_pointer = 0;
	r_t_type.n_type = (EDataType)-1;
	// clear type

	Skip_Space(b, e, r_s_proc);
	int n_begin = b;
	if(!Skip_Identifier(b, e, r_s_proc))
		return false;
	// skip ordinary ident

	const char *p_s_type = r_s_proc.c_str() + n_begin;
	int n_length = b - n_begin;
	// get type string, and it's length

	if(b < e && n_length == 5 && !strncmp(p_s_type, "const", 5) && isspace(p_s_type[n_length])) {
		r_t_type.b_const = true;

		Skip_Space(b, e, r_s_proc);
		n_begin = b;
		if(!Skip_Identifier(b, e, r_s_proc))
			return false;
		// skip ordinary ident

		p_s_type = r_s_proc.c_str() + n_begin;
		n_length = b - n_begin;
		// get type string, and it's length
	}
	// process "const" keyword

	bool b_unsigned = false;

	if(b < e && n_length == 8 && !strncmp(p_s_type, "unsigned", 8) && isspace(p_s_type[n_length])) {
		b_unsigned = true;

		Skip_Space(b, e, r_s_proc);
		n_begin = b;
		if(!Skip_Identifier(b, e, r_s_proc))
			return false;
		// skip ordinary ident

		p_s_type = r_s_proc.c_str() + n_begin;
		n_length = b - n_begin;
		// get type string, and it's length
	}
	// process "unsigned" keyword

	if(n_length > 2 && p_s_type[0] == 'G' && (p_s_type[1] == 'L' || p_s_type[1] == 'l') // GL_INTEL_texture_scissor contains Glenum instead of GLenum
	   && p_s_type[2] >= 'a' && p_s_type[2] <= 'z') {
		n_length -= 2;
		p_s_type += 2;
		n_begin += 2; // for pointer parsing
	}
	//  GL prepended to type

	Skip_Space(b, e, r_s_proc);
	while(b < e && p_s_type[b - n_begin] == '*') {
		++ r_t_type.n_pointer;
		++ b;
		Skip_Space(b, e, r_s_proc);
		if(b + 5 <= e && !strncmp(&p_s_type[b - n_begin], "const", 5)) {
			b += 5;
			Skip_Space(b, e, r_s_proc);
			r_t_type.b_const = true;
		}
		// skip const ("const void * const * a")
	}
	// handle pointers

	for(size_t i = 0; i < sizeof(p_type_name_table) / sizeof(p_type_name_table[0]); ++ i) {
		if(strlen(p_type_name_table[i].p_s_name) == n_length &&
		   !strncmp(p_type_name_table[i].p_s_name, p_s_type, n_length)) {
			r_t_type.n_type = p_type_name_table[i].n_type;

			if(b_unsigned) {
				switch(r_t_type.n_type) {
				case type_Char:
				case type_CharARB:
					r_t_type.n_type = type_UChar;
					break;
				case type_Byte:
					r_t_type.n_type = type_UByte;
					break;
				case type_Short:
					r_t_type.n_type = type_UShort;
					break;
				case type_Int:
					r_t_type.n_type = type_UInt;
					break;
				case type_Int64:
					r_t_type.n_type = type_UInt64;
					break;
				case type_Int64EXT:
					r_t_type.n_type = type_UInt64EXT;
					break;
				default:
					fprintf(stderr, "error: invalid unsigned type\n");
					return false;
				}
			}
			// handle "C"/C++ unsigned types

			return true;
		}
	}
	// find type in the table


	return false;
	// unknown type
}

inline bool CGLExtension::b_IsTokenChar(char n_char, bool b_first) // lowercase characters accepted as well
{
	if(b_first)
		return (n_char >= 'A' && n_char <= 'Z') || (n_char >= 'a' && n_char <= 'z') || n_char == '_';
	else
		return (n_char >= 'A' && n_char <= 'Z') || (n_char >= 'a' && n_char <= 'z') || (n_char >= '0' && n_char <= '9') || n_char == '_';
}

/*
 *								=== ~CGLExtension ===
 */
