/**
 *	@file GLExtension.h
 *	@brief OpenGL extension parser header
 *	@author -tHE SWINe-
 */

#ifndef __OPENGL_EXTENSION_REGISTRY_PARSER_INCLUDED
#define __OPENGL_EXTENSION_REGISTRY_PARSER_INCLUDED

#include <string.h>
#include <vector>
#include <string>
#include <set>
#include <map>

/**
 *	@brief removes leading and trailing whitespace
 *	@param[in,out] r_s_string is the string to have the whitespace removed
 */
void TrimSpace(std::string &r_s_string);

/**
 *	@brief utility class for comparing "C" strings, intended for use with eg. std::set<const char*>
 */
class CCompareCStr {
public:
	/**
	 *	@brief comapres two strings
	 *	@param[in] a is the first string
	 *	@param[in] b is the second string
	 *	@return Returns true if a is lexicographically less than b, otherwise returns false.
	 */
	inline bool operator ()(const char *a, const char *b) const
	{
		return strcmp(a, b) < 0;
	}
};

/**
 *	@brief data type names
 */
enum EDataType {
	type_Void,
	type_Boolean,
	type_Char,
	type_CharARB,
	type_Byte,
	type_Short,
	type_Int,
	type_Float, type_ClampFloat,
	type_Double, type_ClampDouble,
	type_UChar,
	type_UByte,
	type_UShort,
	type_UInt,
	type_Enum,
	type_Int64,
	type_Int64EXT,
	type_UInt64,
	type_UInt64EXT,
	type_IntPtr,
	type_IntPtrARB,
	type_SizeiPtr,
	type_SizeiPtrARB,
	type_Sizei,
	type_SizeiARB,
	type_BitField,
	type_BitFieldARB,
	type_Handle,
	type_HandleARB,
	type_Sync,
	type_SyncARB,
	type_Half,

	type_Fixed,
	type_ClampFixed,

	type_FirstNativeType, // marker

	type_OpenCLContext,
	type_OpenCLEvent,

	type_Display,
	type_GLXDrawable,
	type_HDC,
	type_WinHandle,
	type_WinBool,
	type_GLXVideoCaptureDeviceNV,
	type_HVIDEOINPUTDEVICENV,
	type_UINT,

	type_LastNativeType, // marker

	type_DEBUGPROCAMD,
	type_DEBUGPROCARB,
	type_vdpauSurfaceNV,

	type_TemplateType
};

/**
 *	@breif token to data-type name translation table 
 */
static const struct {
	const char *p_s_name; /**< @brief data type name as it appears in the source code */
	EDataType n_type; /**< @brief the corresponding data type name */
} p_type_name_table[] = {
	{"void", type_Void},
	{"boolean", type_Boolean},
	{"Boolean", type_Boolean}, // a bug in GL_EXT_vertex_array
	{"char", type_Char},
	{"charARB", type_CharARB},
	{"byte", type_Byte},
	{"short", type_Short},
	{"int", type_Int},
	{"float", type_Float},
	{"clampf", type_ClampFloat},
	{"double", type_Double},
	{"clampd", type_ClampDouble},
	{"ubyte", type_UByte},
	{"uchar", type_UChar},
	{"uchar", type_UChar},
	{"ushort", type_UShort},
	{"uint", type_UInt},
	{"enum", type_Enum},
	{"int64", type_Int64},
	{"int64EXT", type_Int64EXT},
	{"uint64", type_UInt64},
	{"uint64EXT", type_UInt64EXT},
	{"intptr", type_IntPtr},
	{"intptrARB", type_IntPtrARB},
	{"sizeiptr", type_SizeiPtr},
	{"sizeiptrARB", type_SizeiPtrARB},
	{"sizei", type_Sizei},
	{"sizeiARB", type_SizeiARB},
	{"bitfield", type_BitField},
	{"bitfieldARB", type_BitFieldARB},
	{"handle", type_Handle},
	{"handleARB", type_HandleARB},
	{"sync", type_Sync},
	{"syncARB", type_SyncARB},
	{"half", type_Half},

	{"cl_context", type_OpenCLContext},
	{"cl_event", type_OpenCLEvent},

	{"fixed", type_Fixed},
	{"clampx", type_ClampFixed},

	{"Display", type_Display},
	{"GLXDrawable", type_GLXDrawable},
	{"HDC", type_HDC},
	{"HANDLE", type_WinHandle},
	{"HVIDEOOUTPUTDEVICENV", type_WinHandle},
	{"BOOL", type_WinBool},
	{"GLXVideoCaptureDeviceNV", type_GLXVideoCaptureDeviceNV},
	{"HVIDEOINPUTDEVICENV", type_HVIDEOINPUTDEVICENV},
	{"UINT", type_UINT},
	{"vdpauSurfaceNV", type_vdpauSurfaceNV},

	{"DEBUGPROCAMD", type_DEBUGPROCAMD},
	{"DEBUGPROCARB", type_DEBUGPROCARB},
	
	{"T", type_TemplateType}, // special type
	{"TYPE", type_TemplateType} // special type
	//{"", },
};

class CXMLNode;

/**
 *	@brief OpenGL extension parser
 */
class CGLExtension {
protected:
	struct TToken {
		std::string s_name, s_comment;
		uint64_t n_value;
		bool b_decimal_value;
		int n_value_digit_num;
	};

	class CFindTokenByName {
	protected:
		const char *m_p_s_name;

	public:
		CFindTokenByName(const char *p_s_name)
			:m_p_s_name(p_s_name)
		{}

		inline bool operator ()(const TToken &r_t_token) const
		{
			return !r_t_token.s_name.compare(m_p_s_name);
		}
	};

	struct TDataType {
		EDataType n_type;
		bool b_const;
		int n_pointer; // 0 = no pointer, 1 = "*", 2 = "**", ...
	};

	struct TParam {
		TDataType t_data_type;
		std::string s_name;
		std::string s_brief; // documentation
	};

	struct TFunction {
		TDataType t_return_type;
		std::string s_name;
		std::string s_brief; // documentation
		std::vector<TParam> param_list;
	};

	std::vector<TToken> m_token_list;
	std::vector<TFunction> m_function_list;

	std::string m_s_ext_name, m_s_ext_brand;

	std::string m_s_proc_string;
	bool m_b_proc_complete;
	int m_n_proc_completeness;

	TFunction m_t_parsed_function;
	TDataType m_t_parsed_param_data_type;

	std::vector<EDataType> m_parsed_function_variant_list;
	std::string m_parsed_function_name_append; // only if m_parsed_function_variant_list not empty

	mutable std::string m_s_temp_data_type_string; // used by p_s_DataType_Name() and p_s_DataType_Name_Space()
	mutable std::string m_s_temp_fname_string; // used by p_s_Function_Name()
	mutable std::string m_s_temp_tname_string; // used by p_s_Token_Name()

	mutable std::map<std::string, std::string> m_download_map;

public:
	// default constructor; p_s_ext_name is eg. GL_ARB_shading_language_100
	CGLExtension(const char *p_s_ext_name); // throws(std::bad_alloc)

	// parse the extension from a file
	bool Parse_FromFile(const char *p_s_filename,
		const std::set<std::string> &r_deprecated_func_set, bool b_find_documentation,
		std::map<std::string, std::string> &r_fun_docu_url_map,
		const std::map<std::string, std::string> &r_fun_docu_alias_map);

	bool b_NoNewTokens() const;

	bool b_NoNewProceduresOrFunctions() const;

	// mark allocated functions to prevent redefinition
	void MarkAllocatedFunctions(std::set<std::string> &r_function_alloc,
		bool b_force_ext_brand_name, bool b_remove_ext_brand_name) const;

	// mark allocated tokens to prevent redefinition
	void MarkAllocatedTokens(std::set<std::string> &token_alloc,
		bool b_force_ext_brand_name, bool b_remove_ext_brand_name) const;

	// just dump to stdout
	void Dump() const;

	// prints source code for CGL4ExtensionHandler
	void Print_Source(FILE *p_fw, const char *p_s_section_comment,
		bool b_force_ext_brand_name, bool b_remove_ext_brand_name,
		bool b_force_ext_brand_name_in_getprocaddress_string,
		bool b_remove_ext_brand_name_in_getprocaddress_string,
		const std::set<std::string> &r_function_alloc,
		const std::map<std::string, std::string> &r_ext_func_name_translation_table,
		const std::set<std::string> &r_deprecated_func_set,
		const std::map<std::string, std::string> &r_entry_point_name_translation_table) const;

	// prints function declarations for CGL4ExtensionHandler
	void Print_Header2(FILE *p_fw, const char *p_s_section_comment,
		std::map<std::string, std::string> &r_ext_func_name_translation_table) const;

	// prints OpenGL extension functions
	void Print_Header(FILE *p_fw, bool b_want_autocheck, const char *p_s_section_comment,
		bool b_force_ext_brand_name, bool b_remove_ext_brand_name,
		const std::set<std::string> &r_token_alloc,
		const std::set<std::string> &r_function_alloc,
		const std::set<std::string> &r_deprecated_func_set) const;

	// prints statically linked OpenGL functions (used for gl10 / gl11)
	void Print_StaticHeader(FILE *p_fw, bool b_want_autocheck, const char *p_s_section_comment,
		bool b_force_ext_brand_name, bool b_remove_ext_brand_name,
		const std::set<std::string> &r_token_alloc,
		const std::set<std::string> &r_function_alloc,
		const std::set<std::string> &r_deprecated_func_set) const;

protected:
	const char *CGLExtension::p_s_Token_Name(const TToken &r_token,
		bool b_force_brand, bool b_remove_brand) const; // throws(std::bad_alloc)
	void DumpTokens(FILE *p_fw, const std::set<std::string> &r_token_alloc,
		const char *p_s_section_comment, bool b_force_ext_brand_name, bool b_remove_ext_brand_name) const;
	bool Parse_Token(const std::string &r_s_token, const char *p_s_filename); // throws(std::bad_alloc)
	static bool b_Have_ParamNames(const TFunction &r_t_func);
	static bool b_Have_Documentation(const TFunction &r_t_func);
	static bool b_Have_SomeDocumentation(const TFunction &r_t_func);
	static bool Make_DummyParamNames(TFunction &r_t_func);
	static const CXMLNode *p_FindElemWithAttrib(const CXMLNode *p_parent, const char *p_s_node_name,
		const char *p_s_attrib_name, const char *p_s_attrib_value);
	bool Retrieve_Doc_ParamNames(TFunction &r_t_func, bool b_need_doc, bool b_need_paranames,
		std::map<std::string, std::string> &r_fun_docu_url_map,
		const std::map<std::string, std::string> &r_fun_docu_alias_map) const; // throws(std::bad_alloc)
	static void PrintFunctionDoc(const TFunction &r_t_func, FILE *p_fw = stdout, const char *p_s_token_name = 0);
	static inline bool isspace_or_separator(char c);
	static std::string s_InnerText(const CXMLNode *p_node); // throws(std::bad_alloc)
	bool Parse_Procedure(const std::string &_r_s_proc_,
		const char *p_s_filename, const std::set<std::string> &r_deprecated_func_set,
		bool b_find_documentation, std::map<std::string, std::string> &r_fun_docu_url_map,
		const std::map<std::string, std::string> &r_fun_docu_alias_map); // throws(std::bad_alloc)
	void Reset_FunctionParser();
	bool b_FunctionParser_NeedsMore() const;
	static void Skip_Space(int &b, int e, const std::string &r_s_proc);
	static bool Skip_Identifier(int &b, int e, const std::string &r_s_proc);
	const char *p_s_Function_Name(const TFunction &r_func,
		bool b_force_brand, bool b_remove_brand) const; // throws(std::bad_alloc)
	const char *p_s_DataType_Name(TDataType t_type) const;
	const char *p_s_DataType_Name_Space(TDataType t_type) const;
	static bool Skip_DataType(TDataType &r_t_type, int &b, int e, const std::string &r_s_proc);
	static inline bool b_IsTokenChar(char n_char, bool b_first); // lowercase characters accepted as well
};

#endif // __OPENGL_EXTENSION_REGISTRY_PARSER_INCLUDED
